# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

# Quick fix remove later
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).resolve().parent.parent.parent.parent))

import math
import os
import shutil
import tempfile
import typing as T
from typing import Union

import esm
import esm.inverse_folding
import esm.inverse_folding.util
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import openmm  # move this and pdb fixer to struc tools
import pandas as pd
import seaborn as sns
import torch
import torch.nn.functional as F
from biotite.structure.io.pdb import PDBFile
from esm.inverse_folding.multichain_util import (
    _concatenate_coords,
    load_complex_coords,
    score_sequence_in_complex,
)
from esm.inverse_folding.util import CoordBatchConverter
from matplotlib.colors import LinearSegmentedColormap

alphabet = torch.load(os.path.join(Path(__file__).parent, "alphabet.pt"))


def esm_compute(
    seqs: list,
    names: list = None,
    model: Union[str, torch.nn.Module] = "esm1v",
    rep_layer: int = 33,
    device=None,
):
    """
    Compute the of esm_tools models for a list of sequences.

    Args:
        seqs (list): protein sequences either as str or biotite.sequence.ProteinSequence.
        names (list, default None): list of names/labels for protein sequences.
            If None sequences will be named seq1, seq2, ...
        model (str, torch.nn.Module): choose either esm2, esm1v or a pretrained model object.
        rep_layer (int): choose representation layer. Default 33.
        device (str): Choose hardware for computation. Default 'None' for autoselection
                          other options are 'cpu' and 'cuda'.

    Returns: representations (list) of sequence representation, batch lens and batch labels

    Example:
        seqs = ["AGAVCTGAKLI", "AGHRFLIKLKI"]
        results, batch_lens, batch_labels = esm_compute(seqs)
    """
    # detect device
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(device)

    # on M1 if mps available
    # if device == torch.device(type='cpu'):
    #    device = torch.device('mps' if torch.backends.mps.is_available() else 'cpu')

    # load model
    if isinstance(model, str):
        if model == "esm2":
            model, alphabet = esm.pretrained.esm2_t33_650M_UR50D()
        elif model == "esm1v":
            model, alphabet = esm.pretrained.esm1v_t33_650M_UR90S()
        else:
            raise ValueError(f"{model} is not a valid model")
    elif isinstance(model, torch.nn.Module):
        alphabet = torch.load(os.path.join(Path(__file__).parent, "alphabet.pt"))
    else:
        raise TypeError("Model should be either a string or a torch.nn.Module object")

    batch_converter = alphabet.get_batch_converter()
    model.eval()
    model.to(device)

    if names is None:
        names = [f"seq{i}" for i in range(len(seqs))]

    data = list(zip(names, seqs))

    # check datatype of sequences - str or biotite
    if all(isinstance(x[0], str) and isinstance(x[1], str) for x in data):
        pass  # all elements are strings
    else:
        data = [(x[0], str(x[1])) for x in data]

    batch_labels, batch_strs, batch_tokens = batch_converter(data)
    batch_lens = (batch_tokens != alphabet.padding_idx).sum(1)

    # Extract per-residue representations (on CPU)
    with torch.no_grad():
        results = model(
            batch_tokens.to(device), repr_layers=[rep_layer], return_contacts=True
        )

    return results, batch_lens, batch_labels, alphabet


def get_seq_rep(results, batch_lens):
    """
    Get sequence representations from esm_compute
    """
    token_representations = results["representations"][33]

    # Generate per-sequence representations via averaging
    sequence_representations = []
    for i, tokens_len in enumerate(batch_lens):
        sequence_representations.append(
            token_representations[i, 1 : tokens_len - 1].mean(0)
        )

    return sequence_representations


def get_logits(results):
    """
    Get logits from esm_compute
    """
    logits = results["logits"]
    return logits


def get_attentions(results):
    """
    Get attentions from esm_compute
    """
    attn = results["attentions"]
    return attn


def get_probability_distribution(logits):
    """
    Convert logits to probability distribution for each position in the sequence.
    """
    # Apply softmax function to the logits along the alphabet dimension (dim=2)
    probability_distribution = F.softmax(logits, dim=-1)

    return probability_distribution


def per_position_entropy(probability_distribution):
    """
    Compute the per-position entropy from a probability distribution tensor.
    """
    # Calculate per-position entropy using the formula: H(x) = -sum(p(x) * log2(p(x)))
    entropy = -torch.sum(
        probability_distribution * torch.log2(probability_distribution + 1e-9), dim=-1
    )
    return entropy


def batch_compute(
    seqs: list = None,
    names: list = None,
    fasta_path: str = None,
    dest: str = None,
    model: str = "esm2",
    batch_size: int = 10,
    rep_layer: int = 33,
    pbar=None,
    device=None,
):
    """
    Computes and saves sequence representations in batches using esm2 or esm1v.

    Args:
        seqs (list): protein sequences either as str or biotite.sequence.ProteinSequence
        names (list, default None): list of names/labels for protein sequences
        fasta_path (str): path to fasta file.
        dest (str): destination where embeddings are saved. Default None (won't save if dest is None).
        model (str): choose either esm2 or esm1v
        batch_size (int): batch size. Default 10
        rep_layer (int): choose representation layer. Default 33.
        pbar: Progress bar for shiny app
        device (str): Choose hardware for computation. Default 'None' for autoselection
                          other options are 'cpu' and 'cuda'.

    Returns: representations (list) of sequence representation.

    Example:
        1.
        seqs = ["AGAVCTGAKLI", "AGHRFLIKLKI"]
        batch_compute(seqs=seqs, dest='path')

        2.
        batch_compute(fasta_path='file.fasta', dest='path')
    """
    if dest is None:
        raise "No save destination provided."

    if fasta_path is None and seqs is None:
        raise "Either fasta_path or seqs must not be None"

    counter = 0
    for i in range(0, len(seqs), batch_size):
        results, batch_lens, _, _ = esm_compute(
            seqs[i : i + batch_size],
            names[i : i + batch_size],
            model=model,
            rep_layer=rep_layer,
            device=device,
        )
        sequence_representations = get_seq_rep(results, batch_lens)
        if dest is not None:
            for j in range(len(sequence_representations)):
                _dest = os.path.join(dest, names[i : i + batch_size][j])
                torch.save(sequence_representations[j], _dest + ".pt")
        if pbar:
            counter += len(seqs[i : i + batch_size])
            pbar.set(
                counter,
                message="Computing",
                detail=f"{counter}/{len(seqs)} computed...",
            )


def mask_positions(sequence: str, mask_char: str = "<mask>"):
    """
    Mask every position of an amino acid sequence. Returns list of masked sequence:

    Args:
        sequence (str): Amino acid sequence
        mask_char (str): Character used for masking (default: <mask>)

    Returns:
        list: list of masked sequences

    Examples:
        sequence = 'AMGAT'
        seqs = mask_positions(sequence)
        ['<mask>MGAT', 'A<mask>GAT', ..., 'AMGA<mask>']
    """
    masked_sequences = []
    for i in range(len(sequence)):
        masked_seq = sequence[:i] + mask_char + sequence[i + 1 :]
        masked_sequences.append(masked_seq)

    return masked_sequences


def get_mutant_logits(
    seq: str,
    model: str = "esm1v",
    batch_size: int = 10,
    rep_layer: int = 33,
    alphabet_size: int = 33,
    pbar=None,
    device=None,
):
    """
    Exhaustively compute the logits for every position in a sequence using esm1v or esm2.
    Every position of a sequence will be masked and the logits for the masked position
    will be calculated. The probabilities for every position will be concatenated in a
    combined logits tensor, which will be returned together with the alphabet.
    Keep in mind: the returned logits have start of sequence and end of sequence tokens attached

    Args:
        seq (str): native protein sequence
        model (str): choose either esm2 or esm1v
        batch_size (int): batch size. Default 10
        rep_layer (int): choose representation layer. Default 33.
        pbar: ProteusAI progress bar.
        device (str): Choose hardware for computation. Default 'None' for autoselection
                          other options are 'cpu' and 'cuda'.

    Returns:
        tuple: torch.Tensor (1, sequence_length, alphabet_size) and alphabet esm_tools.data_tools.Alphabet

    Example:
        1.
        seq = "AGHRFLIKLKI"
        logits = get_mutant_logits(seq=seq)
    """
    assert isinstance(seq, str)

    masked_seqs = mask_positions(
        seq
    )  # list of sequences where every position has been masked once
    names = [f"seq{i}" for i in range(len(masked_seqs))]
    sequence_length = len(seq)

    # Initialize an empty tensor of the desired shape
    logits_tensor = torch.zeros(1, sequence_length, alphabet_size)

    counter = 0
    for i in range(0, len(masked_seqs), batch_size):
        results, batch_lens, batch_labels, alphabet = esm_compute(
            masked_seqs[i : i + batch_size],
            names[i : i + batch_size],
            model=model,
            rep_layer=rep_layer,
            device=device,
        )
        logits = results["logits"]

        counter += len(masked_seqs[i : i + batch_size])

        if pbar:
            pbar.set(
                counter,
                message="Computing",
                detail=f"{counter}/{len(masked_seqs)} positions computed...",
            )
        else:
            print(f"{counter}/{len(masked_seqs)} positions computed...")

        # Extract the logits corresponding to the masked position for each sequence in the batch
        for j, masked_seq_name in enumerate(names[i : i + batch_size]):
            masked_position = int(masked_seq_name[3:])
            logits_tensor[0, masked_position] = logits[j, masked_position + 1]

    return logits_tensor, alphabet


def masked_marginal_probability(
    p: torch.Tensor, wt_seq: str, alphabet: esm.data.Alphabet
):
    """
    Get the masked marginal probabilities for every mutation vs the wildtype.

    Args:
        p (torch.Tensor): probability distribution.
        wt_seq (str): wildtype sequence
        alphabet (esm.data.Alphabet): alphabet

    Returns:
        torch.Tensor: masked marginal probability for every position

    Examples:
        seq = 'SEQVENCE'
        logits, alphabet = get_mutant_logits(seq)
        p = get_probability_distribution(logits)
        mmp = masked_marginal_probability(p, logits, alphabet)
    """

    assert isinstance(wt_seq, str)

    if p.shape[1] == len(wt_seq):
        assert p.shape[1] == len(wt_seq)
    else:
        p = p[:, 1:-1]
        assert p.shape[1] == len(wt_seq)

    if isinstance(alphabet, dict):
        pass
    else:
        try:
            alphabet = alphabet.to_dict()
        except Exception as e:
            raise ValueError(f"alphabet has an unexpected format. Error {e}")

    # Create tensors for the wildtype sequence and canonical amino acid probabilities
    wt_tensor = torch.zeros(1, len(wt_seq), len(alphabet))

    # Populate the tensors with the wildtype logits and canonical probabilities
    for i, aa in enumerate(wt_seq):
        wt_ind = alphabet[aa]
        wt_tensor[0, i] = p[0, i, wt_ind]

    # Compute the masked marginal probabilities
    mmp = torch.log(p) - torch.log(wt_tensor)

    return mmp


def most_likely_sequence(log_prob_tensor, alphabet):
    """
    Get the most likely amino acid sequence based on log probabilities.

    Args:
        log_prob_tensor (torch.Tensor): Tensor of shape (1, sequence_length, alphabet_size) containing log probabilities
        alphabet (dict or esm.data.Alphabet): Dictionary mapping indices to characters

    Returns:
        str: Most likely amino acid sequence
    """
    if isinstance(alphabet, dict):
        pass
    else:
        try:
            alphabet = alphabet.to_dict()
        except Exception as e:
            raise ValueError(f"alphabet has an unexpected format. Error {e}")

    # Find the indices of the maximum log probabilities along the alphabet dimension
    max_indices = torch.argmax(log_prob_tensor, dim=-1).squeeze()

    # Filter the alphabet dictionary to only include canonical AAs
    include = [
        "A",
        "R",
        "N",
        "D",
        "C",
        "Q",
        "E",
        "G",
        "H",
        "I",
        "L",
        "K",
        "M",
        "F",
        "P",
        "S",
        "T",
        "W",
        "Y",
        "V",
    ]
    filtered_alphabet = {i: char for char, i in alphabet.items() if char in include}

    # Map the indices back to their corresponding amino acids using the filtered_alphabet dictionary
    most_likely_seq = "".join(
        [
            filtered_alphabet[int(idx)]
            for idx in max_indices
            if int(idx) in filtered_alphabet
        ]
    )

    return most_likely_seq


def find_mutations(native_seq, predicted_seq):
    """
    Find the mutations between the native protein sequence and the predicted most likely sequence.

    Args:
        native_seq (str): Native protein sequence
        predicted_seq (str): Predicted most likely protein sequence

    Returns:
        list: List of mutations in the format ['G2A', 'F4H']
    """

    if len(native_seq) != len(predicted_seq):
        raise ValueError("Native and predicted sequences must have the same length")

    mutations = []

    for i, (native_aa, predicted_aa) in enumerate(zip(native_seq, predicted_seq)):
        if native_aa != predicted_aa:
            mutation = f"{native_aa}{i+1}{predicted_aa}"
            mutations.append(mutation)

    return mutations


def zs_to_csv(
    wt_seq: str,
    alphabet: esm.data.Alphabet,
    p: torch.Tensor,
    mmp: torch.Tensor,
    entropy: torch.Tensor,
    dest: str,
):
    """
    Save the results as a CSV file.

    Args:
        wt_seq (str): Wildtype sequence.
        alphabet (esm.data.Alphabet): Alphabet used for the model.
        p (torch.Tensor): Probability distribution tensor.
        mmp (torch.Tensor): Masked marginal probability tensor.
        entropy (torch.Tensor): Per-position entropy tensor.
        dest (str): Destination path for the CSV file.
    """
    # Convert alphabet to list for indexing
    alphabet_list = list(alphabet.to_dict().keys())  # noqa: F841
    alphabet = alphabet.to_dict()
    canonical_aas = [
        "A",
        "R",
        "N",
        "D",
        "C",
        "Q",
        "E",
        "G",
        "H",
        "I",
        "L",
        "K",
        "M",
        "F",
        "P",
        "S",
        "T",
        "W",
        "Y",
        "V",
    ]

    mutants, sequences, p_values, mmp_values, entropy_values = [], [], [], [], []
    for pos in range(len(wt_seq)):
        for aa in canonical_aas:
            if wt_seq[pos] != aa:
                mutants.append(wt_seq[pos] + str(pos + 1) + aa)
                sequences.append(wt_seq[:pos] + aa + wt_seq[pos + 1 :])
                p_values.append(p[0, pos, alphabet[aa]].item())
                mmp_values.append(mmp[0, pos, alphabet[aa]].item())
                entropy_values.append(entropy[0, pos].item())

    df = pd.DataFrame(
        {
            "mutant": mutants,
            "sequence": sequences,
            "p": p_values,
            "mmp": mmp_values,
            "entropy": entropy_values,
        }
    )
    df.to_csv(dest, index=False)
    return df


### Protein structure
def string_to_tempfile(data):
    """
    Take a string and return a temporary file object with string as content
    """
    # create a temporary file
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        # write the string to the file
        temp_file.write(data.encode("utf-8"))
        # flush the file to make sure the data_tools is written
        temp_file.flush()
        # return the file object
        return temp_file


def save_tempfile(temp_file, target_file_path):
    """
    Copies the contents of a temporary file to a permanent file location.

    Args:
        temp_file_path (str): The file path of the temporary file.
        target_file_path (str): The file path where the contents should be permanently saved.

    Returns:
        bool: True if the file was copied successfully, False if an error occurred.
    """
    if isinstance(temp_file, tempfile._TemporaryFileWrapper):
        temp_file_path = temp_file.file.name
    elif isinstance(temp_file, str):
        temp_file_path = temp_file
    else:
        raise ValueError("tempfile has an unexpected format")
    try:
        # Copy the content of the temporary file to the target file
        shutil.copy(temp_file_path, target_file_path)
        print(f"File has been successfully saved to {target_file_path}.")
        return True
    except Exception as e:
        print(f"Failed to save the file: {e}")
        return False


def tempfile_to_string(temp_file):
    """
    Take a temporary file object and return its content as a string
    """
    with open(temp_file.name, "r", encoding="utf-8") as file:
        data = file.read()
    return data


def clean_pdb_with_pdbfixer(pdbfile):
    try:
        from pdbfixer import PDBFixer
    except Exception as e:
        raise ValueError(
            f"Cleaning PDB structures requires PDBFixer: Please install through conda:\nconda install conda-forge::pdbfixer. Error {e}"
        )

    fixer = PDBFixer(filename=pdbfile)
    fixer.findMissingResidues()
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.removeHeterogens(keepWater=True)
    fixer.addMissingHydrogens()

    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".pdb") as temp_file:
        openmm.app.PDBFile.writeFile(fixer.topology, fixer.positions, temp_file)
        return temp_file.name


def esm_design(
    pdbfile,
    target_chain,
    chains,
    fixed=[],
    temperature=1.0,
    num_samples=100,
    model=None,
    alphabet=None,
    noise=0.2,
    pbar=None,
):  # TODO: change chains to target chain id
    """
    Perform structure-based protein design for a specific chain within a protein complex using ESM-IF.

    Args:
        pdbfile (str): Path to pdb file.
        target_chain (str): The chain to be designed.
        chains (str): All chains to be considered.
        fixed (list): List of residue indices in the target chain that should remain fixed.
        temperature (float): Sampling temperature. Higher temperatures lead to more stochasticity.
        num_samples (int): Number of samples.
        model (esm.pretrained.esm_if1_gvp4_t16_142M_UR50): If None, model will be loaded.
        alphabet (esm.data.Alphabet): If None, alphabet will be loaded.
        noise (float): Add Gaussian noise to coordinates. Default is 0.2 angstroms.
        pbar: Progress bar for shiny app.

    Returns:
        DataFrame with columns: seqid, recovery, log_likelihood, sequence
    """

    # Load model and alphabet if not provided
    if model is None or alphabet is None:
        model, alphabet = esm.pretrained.esm_if1_gvp4_t16_142M_UR50()

    # Clean the PDB file and load coordinates for all chains in the complex
    cleaned_pdbfile = clean_pdb_with_pdbfixer(pdbfile)
    coords_dict, seqs_dict = load_complex_coords(cleaned_pdbfile, chains)

    # Add noise to the coordinates if specified
    if noise is not None:
        for chain_id in coords_dict:
            coord_noise = np.random.normal(
                0, noise, coords_dict[chain_id].shape
            ).astype(coords_dict[chain_id].dtype)
            coords_dict[chain_id] += coord_noise

    # Extract sequence of the target chain
    target_chain_seq = seqs_dict[target_chain]
    print(
        f"Native sequence for chain {chains} loaded from structure file:\n{target_chain_seq}"
    )

    # Concatenate coordinates with padding
    all_coords = _concatenate_coords(coords_dict, target_chain)
    L = all_coords.shape[0]  # Total length after concatenation
    target_chain_length = len(coords_dict[target_chain])

    # Batch converter for converting to model inputs
    batch_converter = CoordBatchConverter(model.decoder.dictionary)
    batch_coords, confidence, _, _, padding_mask = batch_converter(
        [(all_coords, None, None)], device="cpu"
    )

    # Initialize token array with mask tokens for sampling
    mask_idx = model.decoder.dictionary.get_idx("<mask>")
    sampled_tokens = torch.full((1, 1 + L), mask_idx, dtype=int)
    sampled_tokens[0, 0] = model.decoder.dictionary.get_idx("<cath>")

    # Unmask fixed residues in the target chain
    if fixed:
        for i in fixed:
            res = target_chain_seq[i - 1]  # Adjust to target chain sequence
            sampled_tokens[0, i] = model.decoder.dictionary.get_idx(res)

    # Encoder run (only once) for efficiency
    incremental_state = dict()
    encoder_out = model.encoder(batch_coords, padding_mask, confidence)

    # Initialize storage for sampled sequences
    all_seqs = []
    sampled_tokens_tensor = (
        sampled_tokens.unsqueeze(-1).expand(-1, -1, num_samples).clone()
    )

    # Autoregressive decoding loop for each position of the target chain
    for i in range(1, target_chain_length + 1):
        logits, _ = model.decoder(
            sampled_tokens[:, :i],
            encoder_out,
            incremental_state=incremental_state,
        )
        logits = logits[0].transpose(0, 1)
        logits /= temperature
        probs = F.softmax(logits, dim=-1)
        if sampled_tokens[0, i] == mask_idx:
            sampled_tokens[:, i] = torch.multinomial(probs, 1).squeeze(-1)
            sampled_tokens_tensor[:, i, :] = torch.multinomial(
                probs, num_samples, replacement=True
            ).squeeze(-1)

    # Remove the prepend token and trim to target chain
    sampled_tokens_tensor = sampled_tokens_tensor[
        0, 1 : target_chain_length + 1, :
    ]  # TODO: Ensure the correct tokens are being extracted

    # Collect results
    results = []
    for i in range(num_samples):
        sampled_seq_i = sampled_tokens_tensor[
            :, i
        ]  # TODO: Check if those are the correct tokens for the chain
        sampled_seq = "".join(
            [model.decoder.dictionary.get_tok(a) for a in sampled_seq_i]
        )
        all_seqs.append(sampled_seq)

        # Calculate recovery and log-likelihood for each sampled sequence
        recovery = np.mean([(a == b) for a, b in zip(target_chain_seq, sampled_seq)])
        ll, _ = score_sequence_in_complex(
            model, alphabet, coords_dict, target_chain, sampled_seq
        )  # TODO: Check if scoring is correct

        # Update progress bar if available
        if pbar:
            pbar.set(i, message="Computing", detail=f"{i+1}/{num_samples} remaining...")

        # Append to results
        results.append(
            {
                "names": f"sampled_seq_{i+1}",
                "recovery": recovery,
                "log_likelihood": ll,
                "sequence": sampled_seq,
            }
        )

    # Convert results to a DataFrame
    df = pd.DataFrame(results)
    return df


# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
def create_batched_sequence_datasest(
    sequences: T.List[T.Tuple[str, str]], max_tokens_per_batch: int = 1024
) -> T.Generator[T.Tuple[T.List[str], T.List[str]], None, None]:
    """
    Code taken from https://github.com/facebookresearch/esm/blob/main/scripts/esmfold_inference.py
    """
    batch_headers, batch_sequences, num_tokens = [], [], 0
    for header, seq in sequences:
        if (len(seq) + num_tokens > max_tokens_per_batch) and num_tokens > 0:
            yield batch_headers, batch_sequences
            batch_headers, batch_sequences, num_tokens = [], [], 0
        batch_headers.append(header)
        batch_sequences.append(seq)
        num_tokens += len(seq)

    yield batch_headers, batch_sequences


def structure_prediction(
    seqs: list,
    names: list = None,
    chunk_size: int = 124,
    max_tokens_per_batch: int = 1024,
    num_recycles: int = None,
    pbar=None,
):
    """
    Predict the structure of proteins. The pdb files are returned as 'biotite.structure.io.pdb.PDBFile' objects.
    They can be written to file with pdb.write(os.path.join(dest, "prot.pdb")).

    Args:
        sequences (list): all sequences for structure prediction
        names (list): names of the sequences
        chunck_size (int): Chunks axial attention computation to reduce memory usage from O(L^2) to O(L). Recommended values: 128, 64, 32.
        max_tokens_per_batch (int): Maximum number of tokens per gpu forward-pass. This will group shorter sequences together.
        num_recycles (int): Number of recycles to run. Defaults to number used in training 4.
        pbar: progress bar for app

    Returns:
        all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs
    """
    if pbar:
        pbar.set(message="Loading model weights")
    model = esm.pretrained.esmfold_v1()
    model = model.eval().cuda()
    model.set_chunk_size(chunk_size)

    if names is None:
        names = [f"seq{i}" for i in range(len(seqs))]

    all_sequences = list(zip(names, seqs))
    batched_sequences = create_batched_sequence_datasest(
        all_sequences, max_tokens_per_batch
    )
    all_headers = []
    all_sequences = []
    all_pdbs = []
    pTMs = []
    mean_pLDDTs = []
    if pbar:
        pbar.set(1, message="Computing", detail=f"1/{len(names)} remaining...")
    for i, (headers, sequences) in enumerate(batched_sequences):
        if pbar:
            pbar.set(
                i + 1, message="Computing", detail=f"{i+1}/{len(names)} remaining..."
            )
        output = model.infer(sequences, num_recycles=num_recycles)
        output = {key: value.cpu() for key, value in output.items()}
        pdbs = model.output_to_pdb(output)
        for header, seq, pdb_string, mean_plddt, ptm in zip(
            headers, sequences, pdbs, output["mean_plddt"], output["ptm"]
        ):
            all_headers.append(header)
            all_sequences.append(seq)
            all_pdbs.append(
                PDBFile.read(string_to_tempfile(pdb_string).name)
            )  # biotite pdb file name
            mean_pLDDTs.append(mean_plddt.item())
            pTMs.append(ptm.item())

    return all_headers, all_sequences, all_pdbs, pTMs, mean_pLDDTs


def format_float(float_value: float, str_len: int = 5, round_val: int = 2):
    """
    Format a float value to a string of length 5, rounded to two decimal places.

    Args:
        float_value (float): The float value to format

    Returns:
        str: The formatted string
    """
    rounded_value = round(float_value, round_val)
    value_str = str(rounded_value)
    formatted_str = value_str.rjust(str_len, " ")
    return formatted_str


def entropy_to_bfactor(pdb, entropy_values, trim=False, alphabet_size=33):
    """
    Convert per-position entropy values to b-factors between 0 and 100.

    Args:
        entropy_values (list or numpy.ndarray): List of entropy values
        trim (bool): If True, remove the start and end of sequence tokens (default: False)

    Returns:
        list: List of scaled b-factors
    """
    if not isinstance(pdb, str):
        try:
            pdb = str(pdb)
        except Exception as e:
            raise ValueError(f"invalid input type for pdb. Error {e}")

    # Remove the start and end of sequence tokens, if requested
    if trim:
        entropy_values = entropy_values[:, 1:-1]

    scaled_entropy = 100 * (1 - entropy_values / math.log2(alphabet_size))
    scaled_entropy_list = scaled_entropy.tolist()

    b_factor_strings = [
        [format(x, ".2f").rjust(6) for x in row] for row in scaled_entropy_list
    ][0]
    lines = []
    id = -1
    count = -1
    for i, line in enumerate(str(pdb).split("\n")):
        if line.startswith("ATOM"):
            res_id = int(line[23:26])
            if res_id != id:
                id = res_id
                count += 1
            line = line[:60] + b_factor_strings[count] + line[66:]
        lines.append(line + "\n")
    pdb = PDBFile.read(string_to_tempfile("".join(lines)).name)
    return pdb


# Visualization
def plot_heatmap(
    p,
    alphabet,
    include="canonical",
    dest=None,
    title: str = None,
    remove_tokens: bool = False,
    show: bool = True,
    color_sheme: str = "b",
    highlight_positions: dict = None,
    section=None,
):
    """
    Plot a heatmap of the probability distribution for each position in the sequence.

    Args:
        p (torch.Tensor): probability_distribution torch.Tensor with shape (1, sequence_length, alphabet_size)
        alphabet (dict or esm.data.Alphabet): Dictionary mapping indices to characters
        include (str or list): List of characters to include in the heatmap (default: canonical, include only canonical amino acids)
        dest (str): Optional path to save the plot as an image file (default: None)
        title (str): Title of the plot (default: None)
        remove_tokens (bool): Remove start of sequence and end of sequence tokens (default: False)
        show (bool): Display plot if True (default: True)
        color_sheme (str): Color scheme for the heatmap ('rwb' for red-white-blue, 'r' for reds, 'b' for blues) (default: 'b')
        highlight_positions (dict): Dictionary specifying positions to highlight with the format {position: residue} (default: None)
        section (tuple): Section which should be shown in the plot - low and high end of sequence to be displayed. Show entire sequence if None (default: None)

    Returns:
        fig (matplotlib.figure.Figure): The created matplotlib figure.
    """

    if isinstance(alphabet, dict):
        pass
    else:
        try:
            alphabet = alphabet.to_dict()
        except Exception as e:
            raise ValueError(f"alphabet has an unexpected format. Error {e}")

    # Convert the probability distribution tensor to a numpy array
    probability_distribution_np = p.cpu().numpy().squeeze()

    # Remove the start and end of sequence tokens
    if remove_tokens:
        probability_distribution_np = probability_distribution_np[1:-1, :]

    # make sure section is in range of sequence
    if section is not None:
        assert section[0] < section[1]

        probability_distribution_np = probability_distribution_np[
            section[0] : section[1], :
        ]

    # If no characters are specified, include only amino acids by default
    if include == "canonical":
        include = [
            "A",
            "R",
            "N",
            "D",
            "C",
            "Q",
            "E",
            "G",
            "H",
            "I",
            "L",
            "K",
            "M",
            "F",
            "P",
            "S",
            "T",
            "W",
            "Y",
            "V",
        ]
    elif include == "all":
        include = alphabet.keys()
    elif isinstance(alphabet, list):
        include = include
    else:
        raise "include must either be 'canonical' 'all' or a list of valid elements"

    # Filter the alphabet dictionary based on the 'include' list
    filtered_alphabet = {char: i for char, i in alphabet.items() if char in include}

    # adjust keys and values
    min_val = min(filtered_alphabet.values())

    # Create a pandas DataFrame with appropriate column and row labels
    df = pd.DataFrame(
        probability_distribution_np[:, list(filtered_alphabet.values())],
        columns=[i for i in filtered_alphabet.keys()],
    )

    # Calculate the symmetric vmin and vmax around zero
    abs_max = max(abs(df.min().min()), abs(df.max().max()))

    # colors
    if color_sheme == "rwb":
        colors = ["red", "white", "blue"]
        cmap = LinearSegmentedColormap.from_list("red_white_blue", colors)
        vmin, vmax = -abs_max, abs_max  # Symmetric range around zero
    elif color_sheme == "r":
        cmap = "Reds"
        vmin, vmax = None, None  # No centering for single color schemes
    else:
        cmap = "Blues"
        vmin, vmax = None, None  # No centering for single color schemes

    # Create a heatmap using seaborn
    fig, ax = plt.subplots(figsize=(12, 6))

    ax = plt.gca()
    sns.heatmap(
        df.T,
        cmap=cmap,
        linewidths=0.5,
        annot=False,
        cbar=True,
        ax=ax,
        vmin=vmin,
        vmax=vmax,
    )

    # Adjust x-axis ticks and labels if 'section' is specified
    if section is not None:
        ax.set_xticks(
            np.arange(
                0, section[1] - section[0] + 1, max(1, (section[1] - section[0]) // 10)
            )
        )
        ax.set_xticklabels(
            np.arange(
                section[0], section[1] + 1, max(1, (section[1] - section[0]) // 10)
            )
        )

    # Highlight the mutated positions with a green box
    if highlight_positions is not None:
        for pos, mutated_residue in highlight_positions.items():
            residue_index = filtered_alphabet[mutated_residue]
            rect = patches.Rectangle(
                (pos, residue_index - min_val),
                1,
                1,
                linewidth=1,
                edgecolor="lime",
                facecolor="none",
            )
            ax.add_patch(rect)

    plt.xlabel("Sequence Position")
    plt.ylabel("Residue")
    if title is None:
        plt.title("Per-Position Probability Distribution Heatmap")
    else:
        plt.title(title)

    # Save the plot to the specified destination, if provided
    if dest is not None:
        plt.savefig(dest, dpi=400, bbox_inches="tight")

    # Show the plot, if the 'show' argument is True
    if show:
        plt.show()

    return fig


def plot_per_position_entropy(
    per_position_entropy: torch.Tensor,
    sequence: str,
    highlight_positions: list = None,
    show: bool = False,
    dest: str = None,
    title: str = None,
    section: tuple = None,
    use_normal_ticks: bool = True,
):
    """
    Plot the per position entropy for a given sequence.

    Args:
        per_position_entropy (torch.Tensor): Tensor of per position entropy values with shape (batch_size, sequence_length).
        sequence (str): Protein sequence.
        highlight_positions (list): List of positions to highlight in red (0-indexed) (default: None).
        show (bool): Display the plot if True (default: False).
        dest (str): Optional path to save the plot as an image file (default: None).
        title (str): Title of plot.
        section (tuple): Section of the sequence to plot (default: None). If None, the entire sequence is plotted.
        use_normal_ticks (bool): If True, use normal numerical ticks for x-axis (default: False).

    Returns:
        matplotlib.figure.Figure: The matplotlib figure object for the plot.
    """

    # Convert the tensor to a numpy array
    per_position_entropy_np = per_position_entropy.cpu().numpy()

    # Check if the length of the per_position_entropy and sequence match
    if per_position_entropy_np.shape[1] != len(sequence):
        raise ValueError(
            "The length of per_position_entropy and sequence must be the same."
        )

    # If a section is specified, adjust the per_position_entropy and sequence accordingly
    if section is not None:
        if section[0] < 0 or section[1] > len(sequence):
            raise ValueError("Section indices are out of range.")
        per_position_entropy_np = per_position_entropy_np[:, section[0] : section[1]]
        sequence = sequence[section[0] : section[1]]

    # Create an array of positions for the x-axis
    positions = np.arange(len(sequence))

    # Dynamic tick interval
    tick_interval = 1 if len(sequence) <= 30 else int(len(sequence) / 20)

    # Create a bar plot of per position entropy
    fig = plt.figure(figsize=(20, 6))

    if highlight_positions is None:
        plt.bar(positions, per_position_entropy_np.squeeze())
    else:
        colors = ["red" if pos in highlight_positions else "blue" for pos in positions]
        plt.bar(positions, per_position_entropy_np.squeeze(), color=colors)

    # Set the x-axis labels
    if use_normal_ticks:
        plt.xticks(positions[::tick_interval])
    else:
        plt.xticks(
            positions[::tick_interval],
            [sequence[i] for i in positions[::tick_interval]],
        )

    # Set the labels and title
    plt.xlabel("Sequence Position")
    plt.ylabel("Per Position Entropy")
    if title is None:
        plt.title("Per Position Entropy of Sequence")
    else:
        plt.title(title)

    if dest is not None:
        plt.savefig(dest, dpi=300)

    # Show the plot
    if show:
        plt.show()

    return fig



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
from biotite.sequence import ProteinSequence
import numpy as np


def load_all_fastas(
    path: str, file_type: str = ".fasta", biotite: bool = False
) -> dict:
    """
    Loads all fasta files from a directory, returns the names/ids and sequences as lists.

    Parameters:
        path (str): path to directory containing fasta files
        file_type (str): some fastas are stored with different file endings. Default '.fasta'.
        biotite (bool): returns sequences as biotite.sequence.ProteinSequence object

    Returns:
        dict: dictionary of file_names and tuple (names, sequences)

    Example:
        results = load_fastas('/path/to/fastas')
    """
    file_names = [f for f in os.listdir(path) if f.endswith(file_type)]
    files = [os.path.join(path, f) for f in file_names if f.endswith(file_type)]

    results = {}
    for i, file in enumerate(files):
        names = []
        sequences = []
        with open(file, "r") as f:
            current_sequence = ""
            for line in f:
                line = line.strip()
                if line.startswith(">"):
                    if current_sequence:
                        sequences.append(current_sequence)
                    names.append(line[1:])
                    current_sequence = ""
                else:
                    current_sequence += line
            if biotite:
                sequences.append(ProteinSequence(current_sequence))
            else:
                sequences.append(current_sequence)

        results[file_names[i]] = (names, sequences)
    return results


def load_fasta(file: str, biotite: bool = False) -> tuple:
    """
    Load all sequences in a fasta file. Returns names and sequences

    Parameters:
        file (str): path to file
        biotite (bool): returns sequences as biotite.sequence.ProteinSequence object

    Returns:
        tuple: two lists containing the names and sequences

    Example:
        names, sequences = load_fastas('example.fasta')
    """

    names = []
    sequences = []
    with open(file, "r") as f:
        current_sequence = ""
        for line in f:
            line = line.strip()
            if line.startswith(">"):
                if current_sequence:
                    sequences.append(current_sequence)
                names.append(line[1:])
                current_sequence = ""
            else:
                current_sequence += line
        if biotite:
            sequences.append(ProteinSequence(current_sequence))
        else:
            sequences.append(current_sequence)

    return names, sequences


def write_fasta(names: list, sequences: list, dest: str = None):
    """
    Takes a list of names and sequences and writes a single
    fasta file containing all the names and sequences. The
    files will be saved at the destination

    Parameters:
        names (list): list of sequence names
        sequences (list): list of sequences
        dest (str): path to output file

    Example:
        write_fasta(names, sequences, './out.fasta')
    """
    assert isinstance(names, list) and isinstance(
        sequences, list
    ), "names and sequences must be type list"
    assert len(names) == len(sequences), "names and sequences must have the same length"

    with open(dest, "w") as f:
        for i in range(len(names)):
            f.writelines(">" + names[i] + "\n")
            if i == len(names) - 1:
                f.writelines(sequences[i])
            else:
                f.writelines(sequences[i] + "\n")


def one_hot_encoding(sequence: str):
    """
    Returns one hot encoding for amino acid sequence. Unknown amino acids will be
    encoded with 0.5 at in entire row.

    Parameters:
    -----------
        sequence (str): Amino acid sequence

    Returns:
    --------
        numpy.ndarray: One hot encoded sequence
    """
    # Define amino acid alphabets and create dictionary
    amino_acids = "ACDEFGHIKLMNPQRSTVWY"
    aa_dict = {aa: i for i, aa in enumerate(amino_acids)}

    # Initialize empty numpy array for one-hot encoding
    seq_one_hot = np.zeros((len(sequence), len(amino_acids)))

    # Convert each amino acid in sequence to one-hot encoding
    for i, aa in enumerate(sequence):
        if aa in aa_dict:
            seq_one_hot[i, aa_dict[aa]] = 1.0
        else:
            # Handle unknown amino acids with a default value of 0.5
            seq_one_hot[i, :] = 0.5

    return seq_one_hot


def blosum_encoding(sequence, matrix="BLOSUM62", canonical=True):
    """
    Returns BLOSUM encoding for amino acid sequence. Unknown amino acids will be
    encoded with 0.5 at in entire row.

    Parameters:
    -----------
        sequence (str): Amino acid sequence
        blosum_matrix_choice (str): Choice of BLOSUM matrix. Can be 'BLOSUM50' or 'BLOSUM62'
        canonical (bool): only use canonical amino acids

    Returns:
    --------
        numpy.ndarray: BLOSUM encoded sequence
    """

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.realpath(__file__))

    ### Amino Acid codes
    alphabet_file = os.path.join(script_dir, "matrices/alphabet")
    alphabet = np.loadtxt(alphabet_file, dtype=str)

    # Define BLOSUM matrices
    _blosum50 = (
        np.loadtxt(os.path.join(script_dir, "matrices/BLOSUM50"), dtype=float)
        .reshape((24, -1))
        .T
    )
    _blosum62 = (
        np.loadtxt(os.path.join(script_dir, "matrices/BLOSUM62"), dtype=float)
        .reshape((24, -1))
        .T
    )

    # Choose BLOSUM matrix
    if matrix == "BLOSUM50":
        matrix = _blosum50
    elif matrix == "BLOSUM62":
        matrix = _blosum62
    else:
        raise ValueError(
            "Invalid BLOSUM matrix choice. Choose 'BLOSUM50' or 'BLOSUM62'."
        )

    blosum_matrix = {}
    for i, letter_1 in enumerate(alphabet):
        if canonical:
            blosum_matrix[letter_1] = matrix[i][:20]
        else:
            blosum_matrix[letter_1] = matrix[i]

    # create empty encoding vector
    encoding = np.zeros((len(sequence), len(blosum_matrix["A"])))

    # Convert each amino acid in sequence to BLOSUM encoding
    for i, aa in enumerate(sequence):
        if aa in alphabet:
            encoding[i, :] = blosum_matrix[aa]
        else:
            # Handle unknown amino acids with a default value of 0.5
            encoding[i, :] = 0.5

    return encoding



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import sys
import proteusAI.visual_tools as vis
import proteusAI.ml_tools.bo_tools as BO
import random
import json
import csv
import torch
import pandas as pd
import gpytorch
import numpy as np
from joblib import dump
from typing import Union
from proteusAI.Library import Library
from proteusAI.ml_tools.torch_tools import GP, predict_gp, computeR2
from sklearn.linear_model import Ridge, RidgeClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.svm import SVC, SVR
from sklearn.model_selection import KFold
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

current_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(current_path, "..")
sys.path.append(root_path)


class Model:
    """
    The Model object allows the user to create machine learning models, using
    Library objects as input.

    Attributes:
        model: Class variable holding the model.
        library (proteusAI.Library): Library object to train a model. Default None.
        model (str): Type of model to create. Default random forrest ('rf').
        x (str): Choose vector representation for 'x'. Default 'ohe'.
        split (str): Choose data split. Default random split.
        k_folds (int): Choose k for k-fold cross validation. Default None.
        grid_search (bool): Performe a grid search. Default 'False'.
        custom_params (dict): Provide a dictionary or json file with custom parameters. Default None.
        custom_model (torch.nn.Module): Provide a custom PyTorch model.
        optim (str): Optimizer for training PyTorch models. Default 'adam'.
        lr (float): Learning rate for training PyTorch models. Default 10e-4.
        seed (int): random seed. Default 21.
        test_true (list): List of true values of the test dataset.
        test_predictions (list): Predicted values of the test dataset.
        test_r2 (float): R-squared value of the model on the test set.
        val_true (list): List of true values of the validation dataset.
        val_predictions (list): Predicted values of the validation dataset.
        val_r2 (float): R-squared value of the model on the validation dataset.
    """

    _sklearn_models = ["rf", "knn", "svm", "ffnn", "ridge"]
    _pt_models = ["gp"]
    _in_memory_representations = ["ohe", "blosum50", "blosum62"]

    def __init__(self, **kwargs):
        """
        Initialize a new model.

        Args:
            model: Contains the trained model.
            library (proteusAI.Library): Library object to train a model. Default None.
            model_type (str): Type of model to create. Default random forrest ('rf').
            x (str): Choose vector representation for 'x'. Default 'ohe'.
            split (str): Choose data split. Default random split.
            k_folds (int): Choose k for k-fold cross validation. Default None.
            grid_search (bool): Performe a grid search. Default 'False'.
            custom_params (dict): Provide a dictionary or json file with custom parameters. Default None.
            custom_model (torch.nn.Module): Provide a custom PyTorch model.
            optim (str): Optimizer for training PyTorch models. Default 'adam'.
            lr (float): Learning rate for training PyTorch models. Default 10e-4.
            seed (int): random seed. Default 21.
        """
        self._model = None
        self.train_data = []
        self.test_data = []
        self.val_data = []
        self.test_true = []
        self.test_predictions = []
        self.val_true = []
        self.val_predictions = []
        self.test_r2 = []
        self.val_r2 = []
        self.rep_path = None
        self.dest = None
        self.y_best = None
        self.out_df = None
        self.search_df = None

        # check for device
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Set attributes using the provided kwargs
        self._set_attributes(**kwargs)

    ### args
    def _set_attributes(self, **kwargs):
        defaults = {
            "library": None,
            "model_type": "rf",
            "x": "ohe",
            "rep_path": None,
            "split": (80, 10, 10),
            "k_folds": None,
            "grid_search": False,
            "custom_params": None,
            "custom_model": None,
            "optim": "adam",
            "lr": 10e-4,
            "seed": None,
            "dest": None,
            "pbar": None,
        }

        # Update defaults with provided keyword arguments
        defaults.update(kwargs)

        for key, value in defaults.items():
            setattr(self, key, value)

    def _update_attributes(self, **kwargs):
        defaults = {
            "library": None,
            "model_type": "rf",
            "x": "ohe",
            "rep_path": None,
            "split": (80, 10, 10),
            "k_folds": None,
            "grid_search": False,
            "custom_params": None,
            "custom_model": None,
            "optim": "adam",
            "lr": 10e-4,
            "seed": None,
            "dest": None,
            "pbar": None,
        }

        # Update defaults with provided keyword arguments
        defaults.update(kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def train(self):
        """
        Train the model.

        Args:
            library (proteusAI.Library): Data for training.
            model_type (str): choose the model type ['rf', 'svm', 'knn', 'ffnn'],
            x (str): choose the representation type ['esm2', 'esm1v', 'ohe', 'blosum50', 'blosum62'].
            rep_path (str): Path to representations. Default None - will extract from library object.
            split (tuple or dict): Choose the split ratio of training, testing and validation data as a tuple. Default (80,10,10).
                                   Alternatively, provide a dictionary of proteins, with the keys 'train', 'test', and 'val', with
                                   list of proteins as values for custom data splitting.
            k_folds (int): Number of folds for cross validation.
            grid_search: Enable grid search.
            custom_params: None. Not implemented yet.
            custom_model: None. Not implemented yet.
            optim (str): Choose optimizer for feed forward neural network. e.g. 'adam'.
            lr (float): Choose a learning rate for feed forward neural networks. e.g. 10e-4.
            seed (int): Choose a random seed. e.g. 42
            pbar: Progress bar for shiny app.
        """
        # Update attributes if new values are provided
        # self._update_attributes(**kwargs)

        # split data
        self.train_data, self.test_data, self.val_data = self.split_data()

        # load model
        self._model = self.model()

        # train
        out = None
        if self.model_type in self._sklearn_models:
            out = self.train_sklearn(rep_path=self.rep_path, pbar=self.pbar)
        elif self.model_type in self._pt_models:
            out = self.train_gp(rep_path=self.rep_path, pbar=self.pbar)
        else:
            raise ValueError(
                f"The training method for '{self.model_type}' models has not been implemented yet"
            )

        return out

    ### Helpers ###
    def split_data(self):
        """
        Split data into train, test, and validation sets.

            1. 'random': Randomly splits data points
            2. 'site': Splits data by sites in the protein,
                such that the same site cannot be found
                in the training, testing and validation set
            3. 'custom': Splits the data according to a custom pattern

        Returns:
            tuple: returns three lists of train, test and validation proteins.
        """

        proteins = self.library.proteins

        if self.seed:
            random.seed(self.seed)

        train_data, test_data, val_data = [], [], []

        if isinstance(self.split, tuple):
            train_ratio, test_ratio, val_ratio = tuple(
                value / sum(self.split) for value in self.split
            )
            train_size = int(train_ratio * len(proteins))
            test_size = int(test_ratio * len(proteins))

            random.shuffle(proteins)

            # Split the data
            train_data = proteins[:train_size]
            test_data = proteins[train_size : train_size + test_size]
            val_data = proteins[train_size + test_size :]

        # custom datasplit
        elif isinstance(self.split, dict):
            train_data = self.split["train"]
            test_data = self.split["test"]
            val_data = self.split["val"]

        # TODO: implement other splitting methods
        else:
            raise ValueError(f"The {self.split} split has not been implemented yet...")

        return train_data, test_data, val_data

    def load_representations(self, proteins: list, rep_path: Union[str, None] = None):
        """
        Loads representations for a list of proteins.

        Args:
            proteins (list): List of proteins.
            rep_path (str): Path to representations. If None, the method assumes the
                library project path and the representation type used for training.

        Returns:
            list: List of representations.
        """
        if self.seed:
            torch.manual_seed(self.seed)
            torch.cuda.manual_seed_all(self.seed)

        reps = self.library.load_representations(rep=self.x, proteins=proteins)

        return reps

    def model(self, **kwargs):
        """
        Load or create model according to user specifications and parameters.

        For a complete list and detailed explanation of model parameters, refer to the scikit-learn documentation.
        """

        model_type = self.model_type
        model = None

        # Define the path for the params.json and model
        if self.dest is not None:
            params_path = f"{self.dest}/params.json"
        else:
            params_path = os.path.join(
                f"{self.library.rep_path}",
                f"../models/{self.model_type}/{self.x}/params.json",
            )
            # Save destination for search_results

        # Check if params.json exists
        if not os.path.exists(params_path):
            os.makedirs(os.path.dirname(params_path), exist_ok=True)
            with open(params_path, "w") as f:
                json.dump(kwargs, f)

        if model_type in self._sklearn_models:
            model_params = kwargs.copy()

            if self.y_type == "class":
                if model_type == "rf":
                    model = RandomForestClassifier(**model_params)
                elif model_type == "svm":
                    model = SVC(**model_params)
                elif model_type == "knn":
                    model = KNeighborsClassifier(**model_params)
                elif model_type == "ridge":  # Added support for Ridge Classification
                    model = RidgeClassifier(**model_params)
            elif self.y_type == "num":
                if model_type == "rf":
                    model = RandomForestRegressor(**model_params)
                elif model_type == "svm":
                    model = SVR(**model_params)
                elif model_type == "knn":
                    model = KNeighborsRegressor(**model_params)
                elif model_type == "ridge":  # Added support for Ridge Regression
                    model = Ridge(**model_params)

            return model

        elif model_type in self._pt_models:
            if self.y_type == "class":
                if model_type == "gp":
                    raise ValueError(
                        f"Model type '{model_type}' has not been implemented yet"
                    )
            elif self.y_type == "num":
                if model_type == "gp":
                    model = "GP_MODEL"

            return model
        else:
            raise ValueError(f"Model type '{model_type}' has not been implemented yet")

    def train_sklearn(self, rep_path, pbar=None):
        """
        Train sklearn models and save the model.

        Args:
            rep_path (str): representation path
            pbar: Progress bar for shiny app.
        """
        assert self._model is not None

        if pbar:
            pbar.set(message="Loading representations", detail="...")

        # This is for representations that are not stored in memory
        train = self.load_representations(self.train_data, rep_path=rep_path)
        test = self.load_representations(self.test_data, rep_path=rep_path)
        val = self.load_representations(self.val_data, rep_path=rep_path)

        # handle representations that are not esm
        if len(train[0].shape) == 2:
            train = [x.view(-1) for x in train]
            test = [x.view(-1) for x in test]
            val = [x.view(-1) for x in val]

        x_train = torch.stack(train).cpu().numpy()
        x_test = torch.stack(test).cpu().numpy()
        x_val = torch.stack(val).cpu().numpy()

        # TODO: For representations that are stored in memory the computation happens here:
        if self.library.pred_data:
            self.y_train = [protein.y_pred for protein in self.train_data]
            self.y_test = [protein.y_pred for protein in self.test_data]
            self.y_val = [protein.y_pred for protein in self.val_data]
        else:
            self.y_train = [protein.y for protein in self.train_data]
            self.y_test = [protein.y for protein in self.test_data]
            self.y_val = [protein.y for protein in self.val_data]

        self.val_names = [protein.name for protein in self.val_data]

        if self.k_folds is None:
            if pbar:
                pbar.set(message="Training {self.model_type}", detail="...")

            # train model
            self._model.fit(x_train, self.y_train)

            # prediction on test set
            self.test_r2 = self._model.score(x_test, self.y_test)
            self.y_test_pred = self._model.predict(x_test)
            self.y_train_pred = self._model.predict(x_train)

            # prediction on validation set
            self.val_r2 = self._model.score(x_val, self.y_val)
            self.y_val_pred = self._model.predict(x_val)
            self.y_train_sigma = [None] * len(self.y_train)
            self.y_val_sigma = [None] * len(self.y_val)
            self.y_test_sigma = [None] * len(self.y_test)

            # Save the model
            if self.dest is not None:
                model_save_path = f"{self.dest}/model.joblib"
                csv_dest = f"{self.dest}"
            else:
                model_save_path = os.path.join(
                    f"{self.library.rep_path}",
                    f"../models/{self.model_type}/{self.x}/model.joblib",
                )
                csv_dest = os.path.join(
                    f"{self.library.rep_path}", f"../models/{self.model_type}/{self.x}"
                )

            os.makedirs(os.path.dirname(model_save_path), exist_ok=True)
            dump(self._model, model_save_path)

            # Add predictions to test proteins
            for i in range(len(test)):
                self.test_data[i].y_pred = self.y_test_pred[i]
                self.test_data[i].y_sigma = self.y_test_sigma[i]

            # Save dataframes
            train_df = self.save_to_csv(
                self.train_data,
                self.y_train,
                self.y_train_pred,
                self.y_train_sigma,
                f"{csv_dest}/train_data.csv",
            )
            test_df = self.save_to_csv(
                self.test_data,
                self.y_test,
                self.y_test_pred,
                self.y_test_sigma,
                f"{csv_dest}/test_data.csv",
            )
            val_df = self.save_to_csv(
                self.val_data,
                self.y_val,
                self.y_val_pred,
                self.y_val_sigma,
                f"{csv_dest}/val_data.csv",
            )

            # Save results to a JSON file
            results = {"test_r2": self.test_r2, "val_r2": self.val_r2}
            with open(f"{csv_dest}/results.json", "w") as f:
                json.dump(results, f)

            # add split information to df
            train_df["split"] = "train"
            test_df["split"] = "test"
            val_df["split"] = "val"

            # Concatenate the DataFrames
            self.out_df = pd.concat([train_df, test_df, val_df], axis=0).reset_index(
                drop=True
            )

            self.y_best = max((max(self.y_train), max(self.y_test), max(self.y_val)))

        # handle ensembles
        else:
            # combine train and test
            self.train_data = self.train_data + self.test_data
            x_train = np.concatenate([x_train, x_test])

            kf = KFold(n_splits=self.k_folds, shuffle=True, random_state=self.seed)
            self.y_train = self.y_train + self.y_test
            fold_results = []
            ensemble = []

            for i, data in enumerate(kf.split(x_train)):
                train_index, test_index = data

                if pbar:
                    pbar.set(
                        message=f"Training {self.model_type} {i+1}/{self.k_folds}",
                        detail="...",
                    )

                x_train_fold, x_test_fold = x_train[train_index], x_train[test_index]
                y_train_fold, y_test_fold = (
                    np.array(self.y_train)[train_index],
                    np.array(self.y_train)[test_index],
                )

                self._model = self.model()
                self._model.fit(x_train_fold, y_train_fold)
                test_r2 = self._model.score(x_test_fold, y_test_fold)
                fold_results.append(test_r2)
                ensemble.append(self._model)

            avg_test_r2 = np.mean(fold_results)

            # Store model ensemble as model
            self._model = ensemble

            # Prediction on validation set
            self.val_data, self.y_val_pred, self.y_val_sigma, self.y_val, _ = (
                self.predict(self.val_data)
            )
            self.train_data, self.y_train_pred, self.y_train_sigma, self.y_train, _ = (
                self.predict(self.train_data)
            )
            self.val_r2 = self.score(self.val_data)

            # Save the model
            if self.dest is not None:
                csv_dest = f"{self.dest}"
            else:
                csv_dest = os.path.join(
                    f"{self.library.rep_path}", f"../models/{self.model_type}/{self.x}"
                )

            for i, model in enumerate(ensemble):
                if self.dest is not None:
                    model_save_path = f"{self.dest}/model_{i}.joblib"
                else:
                    model_save_path = os.path.join(
                        f"{self.library.rep_path}",
                        f"../models/{self.model_type}/{self.x}/model_{i}.joblib",
                    )

                os.makedirs(os.path.dirname(model_save_path), exist_ok=True)
                dump(model, model_save_path)

            # Save the sequences, y-values, and predicted y-values to CSV
            train_df = self.save_to_csv(
                self.train_data,
                self.y_train,
                self.y_train_pred,
                self.y_train_sigma,
                f"{csv_dest}/train_data.csv",
            )
            val_df = self.save_to_csv(
                self.val_data,
                self.y_val,
                self.y_val_pred,
                self.y_val_sigma,
                f"{csv_dest}/val_data.csv",
            )

            # Save results to a JSON file
            results = {
                "k_fold_test_r2": fold_results,
                "avg_test_r2": avg_test_r2,
                "val_r2": self.val_r2,
            }
            with open(f"{csv_dest}/results.json", "w") as f:
                json.dump(results, f)

            # add split information to df
            train_df["split"] = "train"
            val_df["split"] = "val"

            # Concatenate the DataFrames
            self.out_df = pd.concat([train_df, val_df], axis=0).reset_index(drop=True)

            self.y_best = max((max(self.y_train), max(self.y_val)))

        # Add predictions to proteins
        for i in range(len(train)):
            self.train_data[i].y_pred = self.y_train_pred[i]
            self.train_data[i].y_sigma = self.y_train_sigma[i]

        # Add predictions to test proteins
        for i in range(len(val)):
            self.val_data[i].y_pred = self.y_val_pred[i]
            self.val_data[i].y_sigma = self.y_val_sigma[i]

        out = {
            "df": self.out_df,
            "rep_path": self.library.rep_path,
            "struc_path": self.library.struc_path,
            "y_type": self.library.y_type,
            "y_col": "y_true",
            "y_pred_col": "y_predicted",
            "y_sigma_col": "y_sigma",
            "seqs_col": "sequence",
            "names_col": "name",
            "reps": self.library.reps,
            "class_dict": self.library.class_dict,
        }

        print(f"Training completed:\nval_r2:\t{self.val_r2}")

        return out

    def train_gp(
        self,
        rep_path,
        epochs=150,
        initial_lr=0.1,
        final_lr=1e-6,
        decay_rate=0.1,
        pbar=None,
    ):
        """
        Train a Gaussian Process model and save the model.

        Args:
            rep_path (str): representation path
            pbar: Progress bar for shiny app.
        """

        assert self._model is not None

        if pbar:
            pbar.set(message="Loading representations", detail="...")

        # This is for representations that are not stored in memory
        train = self.load_representations(self.train_data, rep_path=rep_path)
        test = self.load_representations(self.test_data, rep_path=rep_path)
        val = self.load_representations(self.val_data, rep_path=rep_path)

        # handle representations that are not esm
        if len(train[0].shape) == 2:
            train = [x.view(-1) for x in train]
            test = [x.view(-1) for x in test]
            val = [x.view(-1) for x in val]

        x_train = torch.stack(train).to(device=self.device)
        x_test = torch.stack(test).to(device=self.device)
        x_val = torch.stack(val).to(device=self.device)

        if self.library.pred_data:
            self.y_train = (
                torch.stack(
                    [torch.Tensor([protein.y_pred]) for protein in self.train_data]
                )
                .view(-1)
                .to(device=self.device)
            )
            self.y_test = (
                torch.stack(
                    [torch.Tensor([protein.y_pred]) for protein in self.test_data]
                )
                .view(-1)
                .to(device=self.device)
            )
            y_val = (
                torch.stack(
                    [torch.Tensor([protein.y_pred]) for protein in self.val_data]
                )
                .view(-1)
                .to(device=self.device)
            )
        else:
            self.y_train = (
                torch.stack([torch.Tensor([protein.y]) for protein in self.train_data])
                .view(-1)
                .to(device=self.device)
            )
            self.y_test = (
                torch.stack([torch.Tensor([protein.y]) for protein in self.test_data])
                .view(-1)
                .to(device=self.device)
            )
            y_val = (
                torch.stack([torch.Tensor([protein.y]) for protein in self.val_data])
                .view(-1)
                .to(device=self.device)
            )

        self.val_names = [protein.name for protein in self.val_data]

        self.likelihood = gpytorch.likelihoods.GaussianLikelihood().to(
            device=self.device
        )
        self._model = GP(x_train, self.y_train, self.likelihood).to(device=self.device)
        # fix_mean = True

        optimizer = torch.optim.Adam(self._model.parameters(), lr=initial_lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=decay_rate)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood, self._model)

        # for param in self._model.named_parameters():
        #    print(param)

        # model.mean_module.constant.data.fill_(1)  # FIX mean to 1
        self._model.train()
        self.likelihood.train()
        prev_loss = float("inf")

        if pbar:
            pbar.set(message=f"Training {self.model_type}", detail="...")

        for _ in range(epochs):
            optimizer.zero_grad()
            output = self._model(x_train)
            loss = -mll(output, self.y_train)
            loss.backward()
            optimizer.step()
            scheduler.step()

            # Check for convergence
            if abs(prev_loss - loss.item()) < 0.0001:
                #    print(f'Convergence reached. Stopping training...')
                break

            prev_loss = loss.item()

        print(f"Training completed. Final loss: {loss.item()}")

        # prediction on train set
        y_train_pred, y_train_sigma = predict_gp(self._model, self.likelihood, x_train)
        self.y_train_pred, self.y_train_sigma = (
            y_train_pred.cpu().numpy(),
            y_train_sigma.cpu().numpy(),
        )

        # prediction on test set
        y_test_pred, y_test_sigma = predict_gp(self._model, self.likelihood, x_test)
        self.test_r2 = computeR2(self.y_test, y_test_pred)
        self.y_test_pred, self.y_test_sigma = (
            y_test_pred.cpu().numpy(),
            y_test_sigma.cpu().numpy(),
        )

        # prediction on validation set
        y_val_pred, y_val_sigma = predict_gp(self._model, self.likelihood, x_val)
        self.val_r2 = computeR2(y_val, y_val_pred)
        self.y_train = self.y_train.cpu().numpy()
        self.y_test_pred, self.y_test_sigma = (
            y_test_pred.cpu().numpy(),
            y_test_sigma.cpu().numpy(),
        )
        self.y_val_pred, self.y_val_sigma = (
            y_val_pred.cpu().numpy(),
            y_val_sigma.cpu().numpy(),
        )

        self.y_val = y_val.cpu().numpy()
        self.y_test = self.y_test.cpu().numpy()

        self.y_best = max((max(self.y_train), max(self.y_test), max(self.y_val)))

        # Add predictions to proteins
        for i in range(len(train)):
            self.train_data[i].y_pred = self.y_train_pred[i].item()
            self.train_data[i].y_sigma = self.y_train_sigma[i].item()

        # Add predictions to test proteins
        for i in range(len(test)):
            self.test_data[i].y_pred = self.y_test_pred[i].item()
            self.test_data[i].y_sigma = self.y_test_sigma[i].item()

        # Add predictions to test proteins
        for i in range(len(val)):
            self.val_data[i].y_pred = self.y_val_pred[i].item()
            self.val_data[i].y_pred = self.y_val_sigma[i].item()

        # Save the model
        if self.dest is not None:
            model_save_path = f"{self.dest}/model.pt"
            csv_dest = self.dest
        else:
            model_save_path = os.path.join(
                f"{self.library.rep_path}",
                f"../models/{self.model_type}/{self.x}/model.pt",
            )
            csv_dest = os.path.join(
                f"{self.library.rep_path}", f"../models/{self.model_type}/{self.x}"
            )

        # Save the model
        os.makedirs(os.path.dirname(model_save_path), exist_ok=True)
        torch.save(self._model.state_dict(), model_save_path)

        # save dataframes
        train_df = self.save_to_csv(
            self.train_data,
            self.y_train,
            self.y_train_pred,
            self.y_train_sigma,
            f"{csv_dest}/train_data.csv",
        )
        test_df = self.save_to_csv(
            self.test_data,
            self.y_test,
            self.y_test_pred,
            self.y_test_sigma,
            f"{csv_dest}/test_data.csv",
        )
        val_df = self.save_to_csv(
            self.val_data,
            self.y_val,
            self.y_val_pred,
            self.y_val_sigma,
            f"{csv_dest}/val_data.csv",
        )

        # Save results to a JSON file
        results = {
            "test_r2": self.test_r2,
            "val_r2": self.val_r2,
        }

        with open(f"{csv_dest}/results.json", "w") as f:
            json.dump(results, f)

        # add split information to df
        train_df["split"] = "train"
        test_df["split"] = "test"
        val_df["split"] = "val"

        # Concatenate the DataFrames
        self.out_df = pd.concat([train_df, test_df, val_df], axis=0).reset_index(
            drop=True
        )

        out = {
            "df": self.out_df,
            "rep_path": self.library.rep_path,
            "struc_path": self.library.struc_path,
            "y_type": self.library.y_type,
            "y_col": "y_true",
            "y_pred_col": "y_predicted",
            "y_sigma_col": "y_sigma",
            "seqs_col": "sequence",
            "names_col": "name",
            "reps": self.library.reps,
            "class_dict": self.library.class_dict,
        }

        return out

    # Save the sequences, y-values, and predicted y-values to CSV
    def save_to_csv(
        self,
        proteins,
        y_values,
        y_pred_values,
        y_sigma_values,
        filename,
        acq_scores=None,
    ):
        # Prepare data for CSV and DataFrame
        data = []
        names = [prot.name for prot in proteins]

        # Determine if acquisition scores are provided
        if acq_scores is not None:
            header = [
                "name",
                "sequence",
                "y_value",
                "y_predicted",
                "y_sigma",
                "acq_score",
            ]  # CSV header with acq_scores
        else:
            header = [
                "name",
                "sequence",
                "y_value",
                "y_predicted",
                "y_sigma",
            ]  # CSV header without acq_scores

        with open(filename, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(header)
            for i, (name, protein, y, y_pred, y_sigma) in enumerate(
                zip(names, proteins, y_values, y_pred_values, y_sigma_values)
            ):
                if acq_scores is not None:
                    row = [name, protein.seq, y, y_pred, y_sigma, acq_scores[i]]
                else:
                    row = [name, protein.seq, y, y_pred, y_sigma]
                writer.writerow(row)
                data.append(row)

        # Create a DataFrame from the collected data
        if acq_scores is not None:
            df = pd.DataFrame(
                data,
                columns=[
                    "name",
                    "sequence",
                    "y_true",
                    "y_predicted",
                    "y_sigma",
                    "acq_score",
                ],
            )
        else:
            df = pd.DataFrame(
                data, columns=["name", "sequence", "y_true", "y_predicted", "y_sigma"]
            )

        return df

    def predict(self, proteins: list, rep_path=None, acq_fn="greedy", batch_size=10000):
        """
        Scores the R-squared value for a list of proteins.

        Args:
            proteins (list): List of proteins to make predictions.
            rep_path (str): Path to representations for proteins in the list.
                If None, the library project path and representation type for training
                will be assumed

        Returns:
            list: Predictions generated by the model.
        """
        if self._model is None:
            raise ValueError("Model is 'None'")

        if acq_fn == "ei":
            acq = BO.EI
        elif acq_fn == "greedy":
            acq = BO.greedy
        elif acq_fn == "ucb":
            acq = BO.UCB
        elif acq_fn == "random":
            acq = BO.random_acquisition

        all_y_pred = []
        all_sigma_pred = []
        all_acq_scores = []

        for i in range(0, len(proteins), batch_size):
            batch_proteins = proteins[i : i + batch_size]
            batch_reps = self.load_representations(batch_proteins, rep_path)

            if len(batch_reps[0].shape) == 2:
                batch_reps = [x.view(-1) for x in batch_reps]

            # GP
            if self.model_type == "gp":
                self.likelihood.eval()
                x = torch.stack(batch_reps).to(device=self.device)
                y_pred, sigma_pred = predict_gp(self._model, self.likelihood, x)
                y_pred = y_pred.cpu().numpy()
                sigma_pred = sigma_pred.cpu().numpy()
                acq_score = acq(y_pred, sigma_pred, self.y_best)

            # Handle ensembles
            elif isinstance(self._model, list):
                ys = []
                for model in self._model:
                    x = torch.stack(batch_reps).cpu().numpy()
                    y_pred = model.predict(x)
                    ys.append(y_pred)

                y_stack = np.stack(ys)
                y_pred = np.mean(y_stack, axis=0)
                sigma_pred = np.std(y_stack, axis=0)
                acq_score = acq(y_pred, sigma_pred, self.y_best)

            # Handle single model
            else:
                x = torch.stack(batch_reps).cpu().numpy()
                y_pred = self._model.predict(x)
                sigma_pred = np.zeros_like(y_pred)
                acq_score = acq(y_pred, sigma_pred, self.y_best)

            all_y_pred.extend(y_pred)
            all_sigma_pred.extend(sigma_pred)
            all_acq_scores.extend(acq_score)

        all_y_pred = np.array(all_y_pred)
        all_sigma_pred = np.array(all_sigma_pred)
        all_acq_scores = np.array(all_acq_scores)

        # Sort acquisition scores and get sorted indices
        sorted_indices = np.argsort(all_acq_scores)[::-1]

        # Sort all lists/arrays by the sorted indices
        val_data = [proteins[i] for i in sorted_indices]
        y_val = [prot.y for prot in val_data]
        y_val_pred = all_y_pred[sorted_indices]
        y_val_sigma = all_sigma_pred[sorted_indices]
        sorted_acq_score = all_acq_scores[sorted_indices]

        for i, prot in enumerate(val_data):
            prot.y_pred = y_val_pred[i]
            prot.y_sigma = y_val_sigma[i]

        return val_data, y_val_pred, y_val_sigma, y_val, sorted_acq_score

    def score(self, proteins: list, rep_path=None):
        """
        Make predictions for a list of proteins.

        Args:
            proteins (list): List of proteins to make predictions.
            rep_path (str): Path to representations for proteins in the list.
                If None, the library project path and representation type for training
                will be assumed

        Returns:
            list: Predictions generated by the model.
        """

        if self._model is None:
            raise ValueError("Model is 'None'")

        reps = self.load_representations(proteins, rep_path)

        if len(reps[0].shape) == 2:
            reps = [x.view(-1) for x in reps]

        x = torch.stack(reps).cpu().numpy()
        y = [protein.y for protein in proteins]

        # ensemble
        ensemble_scores = []
        if isinstance(self._model, list):
            for model in self._model:
                score = model.score(x, y)
                ensemble_scores.append(score)
            ensemble_scores = np.stack(ensemble_scores)
            scores = np.mean(ensemble_scores, axis=0)
        else:
            scores = self._model.score(x, y)

        return scores

    def true_vs_predicted(
        self,
        y_true: list,
        y_pred: list,
        title: Union[str, None] = None,
        x_label: Union[str, None] = None,
        y_label: Union[str, None] = None,
        plot_grid: bool = True,
        file: Union[str, None] = None,
        show_plot: bool = False,
    ):
        """
        Predicts true values versus predicted values.

        Args:
            y_true (list): True y values.
            y_pred (list): Predicted y values.
            title (str): Set the title of the plot.
            x_label (str): Set the x-axis label.
            y_label (str): Set the y-axis label.
            plot_grid (bool): Display a grid in the plot.
            file (str): Choose a file name.
            show_plot (bool): Choose to show the plot.
        """

        if self.dest:
            dest = os.path.join(self.dest, "plots")
        else:
            dest = os.path.join(
                self.library.rep_path, f"../models/{self.model_type}/{self.x}/plots"
            )
        if not os.path.exists(dest):
            os.makedirs(dest)

        fig, ax = vis.plot_predictions_vs_groundtruth(
            y_true, y_pred, title, x_label, y_label, plot_grid, file, show_plot
        )

        return fig, ax

    def search(
        self,
        N=10,
        labels=["all"],
        optim_problem="max",
        method="ga",
        max_eval=10000,
        explore=0.1,
        batch_size=100,
        pbar=None,
        acq_fn="ei",
    ):
        """Search for new mutants or select variants from a set of sequences"""

        if self.y_type == "class":
            out, mask = self._class_search(
                N=N, labels=labels, method=method, max_eval=max_eval, pbar=pbar
            )
            return out, mask
        elif self.y_type == "num":
            out = self._num_search(
                method=method,
                optim_problem=optim_problem,
                max_eval=max_eval,
                explore=explore,
                batch_size=batch_size,
                pbar=pbar,
                acq_fn=acq_fn,
            )
            return out

    def _class_search(
        self,
        N=10,
        optim_problem="max",
        labels=["all"],
        method="ga",
        max_eval=10000,
        pbar=None,
    ):
        """
        Sample diverse sequences and return a mask with 1 for selected indices and 0 for non-selected.

        Args:
            N (int): Number of sequences to be returned.
            optim_problem (float): Minimization or maximization of y-values. Default 'max', alternatively 'min'.
            labels (list): list of labels to sample from. Default ['all'] will sample from all labels.
            method (str): Method used for sampling. Default 'ga' - Genetic Algorithm.
            max_eval (int): Maximum number of evaluations. Default 1000.
            pbar: Progress bar for ProteusAI app.
        """

        class_dict = self.library.class_dict
        full_proteins = self.library.proteins  # Full list of proteins

        if "all" in labels or len(labels) < 1:
            labels = list(class_dict.keys())
            proteins = full_proteins
            full_indices = list(range(len(full_proteins)))  # Indices for all proteins
        else:
            # Filter proteins by label and keep track of their original indices
            proteins, full_indices = zip(
                *[
                    (prot, idx)
                    for idx, prot in enumerate(full_proteins)
                    if class_dict[prot.y] in labels
                ]
            )
            proteins = list(proteins)
            full_indices = list(full_indices)

        vectors = self.load_representations(proteins, rep_path=self.library.rep_path)

        if pbar:
            pbar.set(message=f"Searching {N} diverse sequences", detail="...")

        selected_indices, diversity = BO.simulated_annealing(vectors, N, pbar=pbar)

        # Map selected_indices back to full_protein list using full_indices
        full_selected_indices = [full_indices[i] for i in selected_indices]

        # Create a mask for the full protein list
        mask = np.zeros(len(full_proteins), dtype=int)

        # Set the selected indices in the full list to 1
        mask[full_selected_indices] = 1

        # Select the proteins based on the full_selected_indices
        selected_proteins = [full_proteins[i] for i in full_selected_indices]
        ys = [prot.y for prot in selected_proteins]
        y_pred = [prot.y_pred for prot in selected_proteins]
        y_sigma = [prot.y_sigma for prot in selected_proteins]

        # Save the search results
        if self.dest is not None:
            csv_dest = self.dest
        else:
            csv_dest = os.path.join(
                f"{self.library.rep_path}", f"../models/{self.model_type}/{self.x}"
            )

        self.search_df = self.save_to_csv(
            selected_proteins, ys, y_pred, y_sigma, f"{csv_dest}/search_results.csv"
        )

        out = {
            "df": self.search_df,
            "rep_path": self.library.rep_path,
            "struc_path": self.library.struc_path,
            "y_type": self.library.y_type,
            "y_col": "y_true",
            "y_pred_col": "y_predicted",
            "y_sigma_col": "y_sigma",
            "seqs_col": "sequence",
            "names_col": "name",
            "reps": self.library.reps,
            "class_dict": self.library.class_dict,
        }

        return out, mask

    def _num_search(
        self,
        optim_problem="max",
        method="ga",
        max_eval=10000,
        explore=0.1,
        batch_size=100,
        pbar=None,
        acq_fn="ei",
    ):
        """
        Search for improved mutants.

        Args:
            N (int): Number of sequences to be returned.
            optim_problem (float): Minimization or maximization of y-values. Default 'max', alternatively 'min'.
            labels (list): list of labels to sample from. Default ['all'] will sample from all labels.
            method (str): Method used for sampling. Default 'ga' - Genetic Algorithm.
            pbar: Progress bar for ProteusAI app.
        """
        if pbar:
            pbar.set(message=f"Evaluation {max_eval} sequences", detail="...")

        # Sort proteins based on the optimization problem
        if optim_problem == "max":
            proteins = sorted(
                self.library.proteins, key=lambda prot: prot.y_pred, reverse=True
            )
        elif optim_problem == "min":
            proteins = sorted(
                self.library.proteins, key=lambda prot: prot.y_pred, reverse=False
            )
        else:
            raise ValueError(f"'{optim_problem}' is an invalid optimization problem")

        # Extract y values and compute the mean
        ys = [prot.y for prot in proteins]
        mean_y = np.mean(ys)

        # Get the sequences of the top N proteins that have y > mean_y or y < mean_y based on the optimization problem
        if optim_problem == "max":
            improved_seqs = [prot.seq for prot in proteins if prot.y > mean_y]
        elif optim_problem == "min":
            improved_seqs = [prot.seq for prot in proteins if prot.y < mean_y]

        # Introduce random mutations from the mutations dictionary
        seq_lens = set([len(seq) for seq in improved_seqs])
        if len(seq_lens) > 1:
            # allow all amin acids for each position
            aa_list = [
                "A",
                "C",
                "D",
                "E",
                "F",
                "G",
                "H",
                "I",
                "J",
                "K",
                "L",
                "M",
                "N",
                "P",
                "Q",
                "R",
                "S",
                "T",
                "V",
                "W",
                "Y",
            ]
            mutations = {i: aa_list for i in range(1, max(seq_lens))}
        else:
            mutations = BO.find_mutations(improved_seqs)

        # Save destination for search_results
        if self.dest is not None:
            csv_dest = self.dest
        else:
            csv_dest = os.path.join(
                f"{self.library.rep_path}",
                f"../models/{self.model_type}/{self.x}/predictions",
            )
            os.makedirs(csv_dest, exist_ok=True)

        csv_file = os.path.join(csv_dest, f"{self.model_type}_{self.x}_predictions.csv")
        if os.path.exists(csv_file):
            self.search_df = pd.read_csv(csv_file)

        # results file name
        fname = f"{csv_dest}/{self.model_type}_{self.x}.csv"

        if os.path.exists(os.path.join(csv_dest, fname)):
            self.search_df = pd.read_csv(os.path.join(csv_dest, fname))

        mutant_df = self._mutate(
            proteins, mutations, explore=explore, max_eval=max_eval
        )

        out = {
            "df": mutant_df,
            "rep_path": self.library.rep_path,
            "struc_path": self.library.struc_path,
            "y_type": self.library.y_type,
            "seqs_col": "sequence",
            "y_col": "y_true",
            "y_pred_col": "y_predicted",
            "y_sigma_col": "y_sigma",
            "acq_col": "acq_score",
            "names_col": "name",
            "reps": self.library.reps,
            "class_dict": self.library.class_dict,
        }

        library = Library(user=self.library.user, source=out)

        if self.x not in self._in_memory_representations:
            library.compute(method=self.x, pbar=pbar, batch_size=batch_size)

        val_data, y_pred, y_sigma, y_val, acq_score = self.predict(
            library.proteins, acq_fn=acq_fn
        )

        self.search_df = self.save_to_csv(
            val_data, y_val, y_pred, y_sigma, csv_file, acq_scores=acq_score
        )

        return self.search_df

    def _mutate(self, proteins, mutations, explore=0.1, max_eval=100):
        """
        Propose new mutations

        Args:
            proteins (list): list of proteins
            mutations (dict): dictionary of positions and mutations. Index start at 1
                example: {15:['A', 'L', 'I']}
            exploration (float): Exploration ratio, float between 0 and 1 to control
                the exploratory tendency of the sampling algorithm.
            max_eval (int): maximum number of evaluations before termination.

        Returns:
            pandas dataframe
        """

        if self.search_df is not None and not self.search_df.empty:
            mutated_seqs = self.search_df.sequence.to_list()
            mutated_names = self.search_df.name.to_list()
            y_trues = [None] * len(mutated_seqs)
            y_preds = self.search_df.y_predicted.to_list()
            y_sigmas = self.search_df.y_sigma.to_list()
            acq_scores = self.search_df.acq_score.to_list()

        else:
            mutated_seqs = []
            mutated_names = []
            y_trues = []
            y_preds = []
            y_sigmas = []
            acq_scores = []

        for _ in range(max_eval):
            prot = random.choice(proteins)
            name = prot.name
            seq_list = list(prot.seq)

            if random.random() < explore:
                # Explore: random position and random mutation
                pos = random.randint(0, len(seq_list) - 1)
                mut = random.choice("ACDEFGHIKLMNPQRSTVWY")
                mutated_name = name + f"+{seq_list[pos]}{pos+1}{mut}"
            else:
                # Exploit: use known mutation from the provided mutations dictionary
                try:
                    pos, mut_list = random.choice(list(mutations.items()))
                    pos = pos - 1
                    if pos < len(seq_list):
                        mut = random.choice(mut_list)
                    mutated_name = (
                        name + f"+{seq_list[pos]}{pos+1}{mut}"
                    )  # list is indexed at 0 but mutation descriptions at 1
                except Exception:
                    pos = random.randint(0, len(seq_list) - 1)
                    mut = random.choice("ACDEFGHIKLMNPQRSTVWY")
                    mutated_name = name + f"+{seq_list[pos]}{pos+1}{mut}"

            if (
                seq_list[pos] != mut and mutated_name not in mutated_names
            ):  # Exclude mutations to the same residue
                seq_list[pos] = mut
                mutated_seq = "".join(seq_list)
                mutated_seqs.append(mutated_seq)
                mutated_names.append(mutated_name)
                y_trues.append(None)
                y_preds.append(None)
                y_sigmas.append(None)
                acq_scores.append(None)

        out_df = pd.DataFrame(
            {
                "name": mutated_names,
                "sequence": mutated_seqs,
                "y_true": y_trues,
                "y_predicted": y_preds,
                "y_sigma": y_sigmas,
                "acq_score": acq_scores,
            }
        )

        return out_df

    ### Getters and Setters ###
    # Getter and Setter for library
    @property
    def library(self):
        return self._library

    @library.setter
    def library(self, library):
        self._library = library
        if library is not None:
            self.y_type = library.y_type

    # Getter and Setter for model_type
    @property
    def model_type(self):
        return self._model_type

    @model_type.setter
    def model_type(self, model_type):
        self._model_type = model_type

    # Getter and Setter for x
    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = x

    # Getter and Setter for split
    @property
    def split(self):
        return self._split

    @split.setter
    def split(self, split):
        self._split = split

    # Getter and Setter for k_folds
    @property
    def k_folds(self):
        return self._k_folds

    @k_folds.setter
    def k_folds(self, k_folds):
        self._k_folds = k_folds

    # Getter and Setter for grid_search
    @property
    def grid_search(self):
        return self._grid_search

    @grid_search.setter
    def grid_search(self, grid_search):
        self._grid_search = grid_search

    # Getter and Setter for custom_params
    @property
    def custom_params(self):
        return self._custom_params

    @custom_params.setter
    def custom_params(self, custom_params):
        self._custom_params = custom_params

    # Getter and Setter for custom_model
    @property
    def custom_model(self):
        return self._custom_model

    @custom_model.setter
    def custom_model(self, custom_model):
        self._custom_model = custom_model

    # Getter and Setter for optim
    @property
    def optim(self):
        return self._optim

    @optim.setter
    def optim(self, optim):
        self._optim = optim

    # Getter and Setter for lr
    @property
    def lr(self):
        return self._lr

    @lr.setter
    def lr(self, lr):
        self._lr = lr

    # Getter and Setter for seed
    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, seed):
        self._seed = seed



from setuptools import setup, find_packages

setup(
    author="Jonathan Funk",
    author_email="funk.jonathan21@gmail.com",
    url="https://github.com/jonfunk21/ProteusAI",
    packages=find_packages("src"),
    package_dir={"": "src"},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.8",
)



import torch
import os
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Union
import gpytorch


def one_hot_encoder(sequences, alphabet=None, canonical=True, pbar=None, padding=None):
    """
    Encodes sequences provided an alphabet.

    Parameters:
        sequences (list or str): list of amino acid sequences or a single sequence.
        alphabet (list or None): list of characters in the alphabet or None to load from file.
        canonical (bool): only use canonical amino acids.
        padding (int or None): the length to which all sequences should be padded.
                               If None, no padding beyond the length of the longest sequence.

    Returns:
        torch.Tensor: (number of sequences, padding or maximum sequence length, size of the alphabet) for list input
                      (padding or maximum sequence length, size of the alphabet) for string input
    """
    # Check if sequences is a string
    if isinstance(sequences, str):
        singular = True
        sequences = [sequences]  # Make it a list to use the same code below
    else:
        singular = False

    # Load the alphabet from a file if it's not provided
    if alphabet is None:
        # Get the directory of the current script
        script_dir = os.path.dirname(os.path.realpath(__file__))

        ### Amino Acid codes
        alphabet_file = os.path.join(script_dir, "matrices/alphabet")
        alphabet = np.loadtxt(alphabet_file, dtype=str)

    # If canonical is True, only use the first 20 characters of the alphabet
    if canonical:
        alphabet = alphabet[:20]

    # Create a dictionary to map each character in the alphabet to its index
    alphabet_dict = {char: i for i, char in enumerate(alphabet)}

    # Determine the length to which sequences should be padded
    max_sequence_length = max(len(sequence) for sequence in sequences)
    padded_length = padding if padding is not None else max_sequence_length

    n_sequences = len(sequences)
    alphabet_size = len(alphabet)

    # Create an empty tensor of the right size, with padding length
    tensor = torch.zeros((n_sequences, padded_length, alphabet_size))

    # Fill the tensor
    for i, sequence in enumerate(sequences):
        if pbar:
            pbar.set(
                i, message="Computing", detail=f"{i}/{len(sequences)} remaining..."
            )
        for j, character in enumerate(sequence):
            if j >= padded_length:
                break  # Stop if the sequence length exceeds the padded length
            # Get the index of the character in the alphabet
            char_index = alphabet_dict.get(
                character, -1
            )  # Return -1 if character is not in the alphabet
            if char_index != -1:
                # Set the corresponding element of the tensor to 1
                tensor[i, j, char_index] = 1.0

    # If the input was a string, return a tensor of shape (padded_length, alphabet_size)
    if singular:
        tensor = tensor.squeeze(0)

    return tensor


def blosum_encoding(
    sequences, matrix="BLOSUM62", canonical=True, pbar=None, padding=None
):
    """
    Returns BLOSUM encoding for amino acid sequences. Unknown amino acids will be
    encoded with 0.5 in the entire row.

    Parameters:
    -----------
        sequences (list or str): List of amino acid sequences or a single sequence.
        matrix (str): Choice of BLOSUM matrix. Can be 'BLOSUM50' or 'BLOSUM62'.
        canonical (bool): Only use canonical amino acids.
        padding (int or None): The length to which all sequences should be padded.
                               If None, no padding beyond the length of the longest sequence.
        pbar: Progress bar for shiny app.

    Returns:
    --------
        torch.Tensor: BLOSUM encoded sequence.
    """

    # Check if sequences is a string
    if isinstance(sequences, str):
        singular = True
        sequences = [sequences]  # Make it a list to use the same code below
    else:
        singular = False

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.realpath(__file__))

    # Load the alphabet
    alphabet_file = os.path.join(script_dir, "matrices/alphabet")
    alphabet = np.loadtxt(alphabet_file, dtype=str)

    # Define BLOSUM matrices
    _blosum50 = (
        np.loadtxt(os.path.join(script_dir, "matrices/BLOSUM50"), dtype=float)
        .reshape((24, -1))
        .T
    )
    _blosum62 = (
        np.loadtxt(os.path.join(script_dir, "matrices/BLOSUM62"), dtype=float)
        .reshape((24, -1))
        .T
    )

    # Choose BLOSUM matrix
    if matrix == "BLOSUM50":
        matrix = _blosum50
    elif matrix == "BLOSUM62":
        matrix = _blosum62
    else:
        raise ValueError(
            "Invalid BLOSUM matrix choice. Choose 'BLOSUM50' or 'BLOSUM62'."
        )

    # Create the BLOSUM encoding dictionary
    blosum_matrix = {}
    for i, letter_1 in enumerate(alphabet):
        if canonical:
            blosum_matrix[letter_1] = matrix[i][:20]
        else:
            blosum_matrix[letter_1] = matrix[i]

    # Determine the length to which sequences should be padded
    max_sequence_length = max(len(sequence) for sequence in sequences)
    padded_length = padding if padding is not None else max_sequence_length

    n_sequences = len(sequences)
    alphabet_size = len(blosum_matrix["A"])

    # Create an empty tensor of the right size, with padding length
    tensor = torch.zeros((n_sequences, padded_length, alphabet_size))

    # Convert each amino acid in sequence to BLOSUM encoding
    for i, sequence in enumerate(sequences):
        if pbar:
            pbar.set(
                i, message="Computing", detail=f"{i}/{len(sequences)} remaining..."
            )
        for j, aa in enumerate(sequence):
            if j >= padded_length:
                break  # Stop if the sequence length exceeds the padded length
            if aa in alphabet:
                tensor[i, j, :] = torch.tensor(blosum_matrix[aa])
            else:
                # Handle unknown amino acids with a default value of 0.5
                tensor[i, j, :] = 0.5

    # If the input was a string, return a tensor of shape (padded_length, alphabet_size)
    if singular:
        tensor = tensor.squeeze(0)

    return tensor


def plot_attention(attention: list, layer: int, head: int, seq: Union[str, list]):
    """
    Plot the attention weights for a specific layer and head.

    Args:
        attention (list): List of attention weights from the model
        layer (int): Index of the layer to visualize
        head (int): Index of the head to visualize
        seq (str): Input sequence as a list of tokens
    """

    if isinstance(seq, str):
        seq = [char for char in seq]

    # Get the attention weights for the specified layer and head
    attn_weights = attention[layer][head].detach().cpu().numpy()

    # Create a heatmap using seaborn
    plt.figure(figsize=(10, 10))
    sns.heatmap(attn_weights, xticklabels=seq, yticklabels=seq, cmap="viridis")

    # Set plot title and labels
    plt.title(f"Attention weights - Layer {layer + 1}, Head {head + 1}")
    plt.xlabel("Input tokens")
    plt.ylabel("Output tokens")

    # Show the plot
    plt.show()


class GP(gpytorch.models.ExactGP):
    def __init__(
        self, train_x, train_y, likelihood, fix_mean=False
    ):  # special method: instantiate object
        super(GP, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()  # attribute
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        self.mean_module.constant.data.fill_(1)  # Set the mean value to 1
        if fix_mean:
            self.mean_module.constant.requires_grad_(False)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


def predict_gp(model, likelihood, X):
    model.eval()
    likelihood.eval()

    with torch.no_grad():
        predictions = likelihood(model(X))
        y_pred = predictions.mean
        y_std = predictions.stddev
        # lower, upper = predictions.confidence_region()

    return y_pred, y_std


def computeR2(y_true, y_pred):
    """
    Compute R2-values for to torch tensors.

    Args:
        y_true (torch.Tensor): true y-values
        y_pred (torch.Tensor): predicted y-values
    """
    # Ensure the tensors are 1-dimensional
    if y_true.dim() != 1 or y_pred.dim() != 1:
        raise ValueError("Both y_true and y_pred must be 1-dimensional tensors")

    # Compute the mean of true values
    y_mean = torch.mean(y_true)

    # Compute the total sum of squares (SS_tot)
    ss_tot = torch.sum((y_true - y_mean) ** 2)

    # Compute the residual sum of squares (SS_res)
    ss_res = torch.sum((y_true - y_pred) ** 2)

    # Compute the R² value
    r2 = 1 - (ss_res / ss_tot)

    return r2.item()  # Convert tensor to float



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for mining_tools.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.mining_tools.alphafoldDB import *  # noqa: F403
from proteusAI.mining_tools.blast import *  # noqa: F403
from proteusAI.mining_tools.uniprot import *  # noqa: F403



# This source code is part of the proteusAI package and is distributed
# under the MIT License.

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

import os
import random
import numpy as np
from proteusAI.design_tools import Constraints
import pandas as pd


class ProteinDesign:
    """
    Optimizes a protein sequence based on a custom energy function. Set weights of constraints to 0 which you don't want to use.

    Parameters:
    -----------
        native_seq (str): native sequence to be optimized
        constraints (dict): constraints on sequence.
            Keys describe the kind of constraint and values the position on which they act.
        sampler (str): choose between simulated_annealing and substitution design_tools.
            Default 'simulated annealing'
        n_traj (int): number of independent trajectories per sampling step. Lowest energy mutant will be selected when
            multiple are viable. Default 16
        steps (int): number of sampling steps per trajectory.
            For simulated annealing, the number of iterations is often chosen in the range of [1,000, 10,000].
        T (float): sampling temperature.
            For simulated annealing, T0 is often chosen in the range [1, 100]. default 1
        M (float): rate of temperature decay.
            or simulated annealing, a is often chosen in the range [0.01, 0.1] or [0.001, 0.01]. Default 0.01
        mut_p (tuple): probabilities for substitution, insertion and deletion.
            Default [0.6, 0.2, 0.2]
        pred_struc (bool): if True predict the structure of the protein at every step and use structure
            based constraints in the energy function. Default True.
        max_len (int): maximum length sequence length for lenght constraint.
            Default 300.
        w_len (float): weight of length constraint.
            Default 0.01
        w_identity (float): Weight of sequence identity constraint. Positive values reward low sequence identity to native sequence.
            Default 0.04
        w_ptm (float): weight for ptm. pTM is calculated as 1-pTM, because lower energies should be better.
            Default 1.
        w_plddt (float): weight for plddt.  The mean pLDDT is calculated as 1-mean_pLDDT, because lower energies should be better.
            Default 1.
        w_globularity (float): weight of globularity constraint
            Default 0.001
        w_bb_coord (float): weight on backbone coordination constraint. Constraints backbone to native structure.
            Default 0.02
        w_all_atm (float): weight on all atom coordination constraint. Acts on all atoms which are constrained
            Default 0.15
        w_sasa (float): weight of surface exposed hydrophobics constraint
            Default 0.02
        outdir (str): path to output directory.
            Default None
        verbose (bool): if verbose print information
    """

    def __init__(
        self,
        native_seq: str = None,
        constraints=None,
        sampler: str = "simulated_annealing",
        n_traj: int = 16,
        steps: int = 1000,
        T: float = 10.0,
        M: float = 0.01,
        mut_p: list = (0.6, 0.2, 0.2),
        pred_struc: bool = True,
        max_len: int = 300,
        w_len: float = 0.01,
        w_identity: float = 0.1,
        w_ptm: float = 1,
        w_plddt: float = 1,
        w_globularity: float = 0.001,
        w_bb_coord: float = 0.02,
        w_all_atm: float = 0.15,
        w_sasa: float = 0.02,
        outdir: str = None,
        verbose: bool = False,
    ):

        if constraints is None:
            constraints = {"no_mut": [], "all_atm": []}
        self.native_seq = native_seq
        self.sampler = sampler
        self.n_traj = n_traj
        self.steps = steps
        self.mut_p = mut_p
        self.T = T
        self.M = M
        self.pred_struc = pred_struc
        self.max_len = max_len
        self.w_max_len = w_len
        self.w_identity = w_identity
        self.w_ptm = w_ptm
        self.w_plddt = w_plddt
        self.w_globularity = w_globularity
        self.outdir = outdir
        self.verbose = verbose
        self.constraints = constraints
        self.w_sasa = w_sasa
        self.w_bb_coord = w_bb_coord
        self.w_all_atm = w_all_atm

        # Parameters
        self.ref_pdbs = None
        self.ref_constraints = None
        self.initial_energy = None

    def __str__(self):
        lines: list[str] = [
            "ProteusAI.MCMC.Design class: \n",
            "---------------------------------------\n",
            "When Hallucination.run() sequences will be hallucinated using this seed sequence:\n\n",
            f"{self.native_seq}\n",
            "\nThe following variables were set:\n\n",
            "variable\t|value\n",
            "----------------+-------------------\n",
            f"algorithm: \t|{self.sampler}\n",
            f"steps: \t\t|{self.steps}\n",
            f"n_traj: \t|{self.n_traj}\n",
            f"mut_p: \t\t|{self.mut_p}\n",
            f"T: \t\t|{self.T}\n",
            f"M: \t\t|{self.M}\n\n",
            "The energy function is a linear combination of the following constraints:\n\n",
            "constraint\t|value\t|weight\n",
            "----------------+-------+------------\n",
            f"length \t\t|{self.max_len}\t|{self.w_max_len}\n",
            f"identity\t|\t|{self.w_identity}\n",
        ]
        s = "".join(lines)
        if self.pred_struc:
            lines = [
                s,
                f"pTM\t\t|\t|{self.w_ptm}\n",
                f"pLDDT\t\t|\t|{self.w_plddt}\n",
                f"bb_coord\t\t|{self.w_bb_coord}\n",
                f"all_atm\t\t|\t|{self.w_all_atm}\n",
                f"sasa\t\t|\t|{self.w_sasa}\n",
            ]
            s = "".join(lines)
        return s

    ### SAMPLERS
    def mutate(self, seqs, mut_p: list = None, constraints: list = None):
        """
        mutates input sequences.

        Parameters:
            seqs (list): list of peptide sequences
            mut_p (list): mutation probabilities
            constraints (list): dictionary of constraints

        Returns:
            list: mutated sequences
        """

        if mut_p is None:
            mut_p = [0.6, 0.2, 0.2]

        AAs = (
            "A",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "K",
            "L",
            "M",
            "N",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "V",
            "W",
            "Y",
        )

        mut_types = ("substitution", "insertion", "deletion")

        mutated_seqs = []
        mutated_constraints = []
        mutations = []
        for i, seq in enumerate(seqs):
            mut_constraints = {}

            # loop until allowed mutation has been selected
            mutate = True
            while mutate:
                pos = random.randint(0, len(seq) - 1)
                mut_type = random.choices(mut_types, mut_p)[0]
                if pos in constraints[i]["no_mut"] or pos in constraints[i]["all_atm"]:
                    pass
                # secondary structure constraint disallows deletion
                # insertions between two secondary structure constraints will have the constraint of their neighbors
                else:
                    break

            if mut_type == "substitution":
                replacement = random.choice(AAs)
                mut_seq = "".join([seq[:pos], replacement, seq[pos + 1 :]])
                for const in constraints[i].keys():
                    positions = constraints[i][const]
                    mut_constraints[const] = positions
                mutations.append(f"sub:{seq[pos]}{pos}{replacement}")

            elif mut_type == "insertion":
                insertion = random.choice(AAs)
                mut_seq = "".join([seq[:pos], insertion, seq[pos:]])
                # shift constraints after insertion
                for const in constraints[i].keys():
                    positions = constraints[i][const]
                    positions = [i if i < pos else i + 1 for i in positions]
                    mut_constraints[const] = positions
                mutations.append(f"ins:{pos}{insertion}")

            elif mut_type == "deletion" and len(seq) > 1:
                lines = list(seq)
                del lines[pos]
                mut_seq = "".join(lines)
                # shift constraints after deletion
                for const in constraints[i].keys():
                    positions = constraints[i][const]
                    positions = [i if i < pos else i - 1 for i in positions]
                    mut_constraints[const] = positions
                mutations.append(f"del:{seq[pos]}{pos}")

            else:
                # will perform insertion if length is to small
                insertion = random.choice(AAs)
                mut_seq = "".join([seq[:pos], insertion, seq[pos:]])
                # shift constraints after insertion
                for const in constraints[i].keys():
                    positions = constraints[i][const]
                    positions = [i if i < pos else i + 1 for i in positions]
                    mut_constraints[const] = positions
                mutations.append(f"ins:{pos}{insertion}")

            mutated_seqs.append(mut_seq)
            mutated_constraints.append(mut_constraints)

        return mutated_seqs, mutated_constraints, mutations

    ### ENERGY FUNCTION and ACCEPTANCE CRITERION
    def energy_function(self, seqs: list, i: int, constraints: list):
        """
        Combines constraints into an energy function. The energy function
        returns the energy values of the mutated files and the associated pdb
        files as temporary files. In addition it returns a dictionary of the different
        energies.

        Parameters:
            seqs (list): list of sequences
            i (int): current iteration in sampling
            constraints (list): list of constraints

        Returns:
            tuple: Energy value, pdbs, energy_log
        """
        # reinitialize energy
        energies = np.zeros(len(seqs))
        energy_log = dict()

        e_len = self.w_max_len * Constraints.length_constraint(
            seqs=seqs, max_len=self.max_len
        )
        e_identity = self.w_identity * Constraints.seq_identity(
            seqs=seqs, ref=self.native_seq
        )

        energies += e_len
        energies += e_identity

        energy_log[f"e_len x {self.w_max_len}"] = e_len
        energy_log[f"e_identity x {self.w_identity}"] = e_identity

        pdbs = []
        if self.pred_struc:
            # structure prediction
            names = [f"sequence_{j}_cycle_{i}" for j in range(len(seqs))]
            headers, sequences, pdbs, pTMs, mean_pLDDTs = (
                Constraints.structure_prediction(seqs, names)
            )
            pTMs = [1 - val for val in pTMs]
            mean_pLDDTs = [1 - val / 100 for val in mean_pLDDTs]

            e_pTMs = self.w_ptm * np.array(pTMs)
            e_mean_pLDDTs = self.w_plddt * np.array(mean_pLDDTs)
            e_globularity = self.w_globularity * Constraints.globularity(pdbs)
            e_sasa = self.w_sasa * Constraints.surface_exposed_hydrophobics(pdbs)

            energies += e_pTMs
            energies += e_mean_pLDDTs
            energies += e_globularity
            energies += e_sasa

            energy_log[f"e_pTMs x {self.w_ptm}"] = e_pTMs
            energy_log[f"e_mean_pLDDTs x {self.w_plddt}"] = e_mean_pLDDTs
            energy_log[f"e_globularity x {self.w_globularity}"] = e_globularity
            energy_log[f"e_sasa x {self.w_sasa}"] = e_sasa

            # there are now ref pdbs before the first calculation
            if self.ref_pdbs is not None:
                e_bb_coord = self.w_bb_coord * Constraints.backbone_coordination(
                    pdbs, self.ref_pdbs
                )
                e_all_atm = self.w_all_atm * Constraints.all_atom_coordination(
                    pdbs, self.ref_pdbs, constraints, self.ref_constraints
                )

                energies += e_bb_coord
                energies += e_all_atm

                energy_log[f"e_bb_coord x {self.w_bb_coord}"] = e_bb_coord
                energy_log[f"e_all_atm x {self.w_all_atm}"] = e_all_atm
            else:
                energy_log[f"e_bb_coord x {self.w_bb_coord}"] = []
                energy_log[f"e_all_atm x {self.w_all_atm}"] = []

        energy_log["iteration"] = i + 1

        return energies, pdbs, energy_log

    def p_accept(self, E_x_mut, E_x_i, T, i, M):
        """
        Decides to accep or reject changes. Changes which have a lower energy
        than the previous state will always be accepted. Changes which have
        higher energies will be accepted with a probability p_accept. The
        acceptance probability for bad states decreases over time.

        Parameters:
        -----------
            E_x_mut (np.array): energies of mutated sequences
            E_x_i (np.array): energies of initial sequences
            T (float): Temperature
            i (int): current itteration
            M (float): decay constant

        Returns:
        --------
            np.array: accept probabilities
        """
        T = T / (1 + M * i)
        dE = E_x_i - E_x_mut
        exp_val = np.exp(1 / (T * dE))
        p_accept = np.minimum(exp_val, np.ones_like(exp_val))
        return p_accept

    ### RUN
    def run(self):
        """
        Runs MCMC-sampling based on user defined inputs. Returns optimized sequences.
        """
        native_seq = self.native_seq
        constraints = self.constraints
        n_traj = self.n_traj
        steps = self.steps
        sampler = self.sampler
        energy_function = self.energy_function
        T = self.T
        M = self.M
        p_accept = self.p_accept
        mut_p = self.mut_p
        outdir = self.outdir
        pdb_out = os.path.join(outdir, "pdbs")
        png_out = os.path.join(outdir, "pngs")
        data_out = os.path.join(outdir, "data_tools")

        if outdir is not None:
            if not os.path.exists(outdir):
                os.mkdir(outdir)
            if not os.path.exists(pdb_out):
                os.mkdir(pdb_out)
            if not os.path.exists(png_out):
                os.mkdir(png_out)
            if not os.path.exists(data_out):
                os.mkdir(data_out)

        if sampler == "simulated_annealing":
            mutate = self.mutate

        if native_seq is None:
            raise "The optimizer needs a sequence to run. Define a sequence by calling SequenceOptimizer(native_seq = <your_sequence>)"

        seqs = [native_seq for _ in range(n_traj)]
        constraints = [constraints for _ in range(n_traj)]
        self.ref_constraints = constraints.copy()  # THESE ARE CORRECT

        # for initial calculation don't use the full sequences, unecessary
        # calculation of initial state
        E_x_i, pdbs, energy_log = energy_function([seqs[0]], -1, [constraints[0]])
        E_x_i = [E_x_i[0] for _ in range(n_traj)]
        pdbs = [pdbs[0] for _ in range(n_traj)]

        # empty energies dictionary for the first run
        for key in energy_log.keys():
            energy_log[key] = []
        energy_log["T"] = []
        energy_log["M"] = []
        energy_log["mut"] = []
        energy_log["description"] = []

        self.initial_energy = E_x_i.copy()
        self.ref_pdbs = pdbs.copy()

        if self.pred_struc and outdir is not None:
            # saves the n th structure
            num = "{:0{}d}".format(len(energy_log["iteration"]), len(str(self.steps)))
            pdbs[0].write(os.path.join(pdb_out, f"{num}_design.pdb"))

        # write energy_log in data_out
        if outdir is not None:
            df = pd.DataFrame(energy_log)
            df.to_csv(os.path.join(data_out, "energy_log.pdb"), index=False)

        for i in range(steps):
            mut_seqs, _constraints, mutations = mutate(seqs, mut_p, constraints)
            E_x_mut, pdbs_mut, _energy_log = energy_function(mut_seqs, i, _constraints)
            # accept or reject change
            p = p_accept(E_x_mut, E_x_i, T, i, M)

            new_struc_found = False
            accepted_ind = []  # indices of accepted structures
            for n in range(n_traj):
                if p[n] > random.random():
                    accepted_ind.append(n)
                    E_x_i[n] = E_x_mut[n]
                    seqs[n] = mut_seqs[n]
                    constraints[n] = _constraints[n]
                    new_struc_found = True

            if new_struc_found:
                # get index of lowest energie sructure out of the newly found structures
                min_E = accepted_ind[0]
                for a in accepted_ind:
                    if a < min_E:
                        min_E = a

                # update all to lowest energy structure
                E_x_i = [E_x_i[min_E] for _ in range(n_traj)]
                seqs = [seqs[min_E] for _ in range(n_traj)]
                constraints = [constraints[min_E] for _ in range(n_traj)]
                pdbs = [pdbs_mut[min_E] for _ in range(n_traj)]

                for key in energy_log.keys():
                    # skip skalar values in this step
                    if key not in ["T", "M", "iteration", "mut", "description"]:
                        e = _energy_log[key]
                        energy_log[key].append(e[min_E].item())

                energy_log["iteration"].append(i)
                energy_log["T"].append(T)
                energy_log["M"].append(M)
                energy_log["mut"].append(mutations[min_E])

                num = "{:0{}d}".format(
                    len(energy_log["iteration"]), len(str(self.steps))
                )
                energy_log["description"].append(f"{num}_design")

                if self.pred_struc and outdir is not None:
                    # saves the n th structure
                    pdbs[0].write(os.path.join(pdb_out, f"{num}_design.pdb"))

                # write energy_log in data_out
                if outdir is not None:
                    df = pd.DataFrame(energy_log)
                    df.to_csv(os.path.join(data_out, "energy_log.pdb"), index=False)

        return seqs
